#include "ovcontrol_3d.h"

#ifndef _ADROID
#include "../widgets/ovmenu_3d_filtersettings.h"
#endif
#include "../widgets/ovmenu_3d_palettes.h"
#include <QSettings>
#include "3d/plot_3dplot.h"
#include "3d/plot_3dpure.h"
#include "../../src/gui/controls.h"
#include "qwt_plot_layout.h"
#include "core/plot_core.h"
#include <QAbstractItemView>
#include <QTimer>

namespace plot {

ovControl_3d::ovControl_3d(ThreeDOverlayButtons *overlay)
{
    this->overlay = overlay;
    parentPlot = overlay->parentPlot;
    PlotControl = overlay->PlotControl;
    connect(overlay, &ThreeDOverlayButtons::GUIPlotSetVisible, this, &ovControl_3d::setVisiblePlot);
    connect(overlay, &ThreeDOverlayButtons::GUIPlotResized, this, &ovControl_3d::resized);
    connect(overlay, &ThreeDOverlayButtons::ApplyEnterClicked, this, &ovControl_3d::ApplyEnterClicked);
    connect(overlay, &ThreeDOverlayButtons::BackCancelClicked, this, &ovControl_3d::BackCancelClicked);
    
    #ifndef _ADROID
    filterSettingsMenu = new ovMenu_3d_FilterSettings(overlay);
#endif
    palettesMenu = new ovmenu_3d_palettes(overlay);
    
    {
#ifdef _ADROID
        QSettings settings("Param/gui.conf", QSettings::IniFormat);
#else
        QSettings settings("Param/scan.conf", QSettings::IniFormat);
#endif
        showHelpButton = settings.value("show_help_buttons", true).toBool();
    }
    
    BPointSelection = new QPushButton();
    BPointSelection->setIcon(makeHugeIcon(":/plot/point.png"));
    BPointSelection->setToolTip(QObject::tr("Point selection"));
    BPointSelection->setFlat(true);
    overlay->setButtonStyle(BPointSelection);
    BPointSelection->hide();
    connect(BPointSelection, SIGNAL(released()), this, SLOT(onPointSelection_changed()));

    BMarkers = new QPushButton();
    BMarkers->setIcon(makeHugeIcon(":/plot/cross_section.png"));
    BMarkers->setToolTip(QObject::tr("Markers"));
    BMarkers->setFlat(true);
    overlay->setButtonStyle(BMarkers);
    BMarkers->hide();
    connect(BMarkers, SIGNAL(released()), this, SLOT(onMarkers_changed()));

    InitFilters();
    
    BPalette = new QPushButton();
    BPalette->hide();
    BPalette->setFlat(true);
    overlay->setButtonStyle(BPalette);
    BPalette->setIcon(makeHugeIcon(":/plot/palettePlot.png"));
    BPalette->setToolTip(QObject::tr("Palettes"));
    connect(BPalette, SIGNAL(released()), this, SLOT(colorPickClicked()));

    BAreaZzoom = new QPushButton();
    BAreaZzoom->setIcon(makeHugeIcon(":/plot/Pall_Zoom_Of.png"));
    BAreaZzoom->setToolTip(QObject::tr("Set Z range by selection"));
    BAreaZzoom->setFlat(true);
    overlay->setButtonStyle(BAreaZzoom);
    BAreaZzoom->hide();
    connect(BAreaZzoom, SIGNAL(released()), this, SLOT(on3DBAreaZoom_changed()));

    BSaveImage = new QPushButton();
    BSaveImage->hide();
    BSaveImage->setFlat(true);
    overlay->setButtonStyle(BSaveImage);
    BSaveImage->setIcon(makeHugeIcon(":/plot/picturePlot2D.png"));
    BSaveImage->setToolTip(QObject::tr("Image snapshot"));
    connect(BSaveImage, SIGNAL(released()), this, SLOT(onSaveImage_clicked()));
    
    InitOldFilters();

    CreateActionsAndMenus();

    helpButton = new QPushButton();
    helpButton->setParent(parentPlot);
    helpButton->hide();
    helpButton->setFlat(true);
    overlay->setButtonStyle(helpButton);
    helpButton->setIcon(makeHugeIcon(":/plot/helpBtn.png"));
    helpButton->setToolTip(QObject::tr("Show help"));
    connect(helpButton, SIGNAL(released()), this, SLOT(onHelpBtn_clicked()));
    overlay->allBtnsList.append(helpButton);
    overlay->allWidgets.append(helpButton);

    helpScrollArea = new QScrollArea();
    helpScrollArea->hide();
    helpScrollArea->setStyleSheet("QSplitter {border: 40px solid black; border-style: outset; border-width: 40px; border-radius: 10px; border-color: #B4B4B4; background-color: #E3E3EC;}");
    helpScrollArea->setParent(parentPlot);
    
    connect(MMathProcessing->actions()[0], SIGNAL(toggled(bool)), this, SLOT(on_GwyXFit_triggered(bool)));
    connect(MMathProcessing->actions()[1], SIGNAL(toggled(bool)), this, SLOT(on_GwyYFit_triggered(bool)));

    QString strMean = plot::T3dPlotControl::NameOfMeanValueFilter();
    QString strMedian = plot::T3dPlotControl::NameOfMedianValueFilter();
    QString strGauss = plot::T3dPlotControl::NameOfGaussianFilter();
    QString strHistogram = plot::T3dPlotControl::NameOfHistogramFilter();
    int idMean = plot::T3dPlotControl::GetNumOfAction(strMean);
    int idMedian = plot::T3dPlotControl::GetNumOfAction(strMedian);
    int idGauss = plot::T3dPlotControl::GetNumOfAction(strGauss);
    int idHistogram = plot::T3dPlotControl::GetNumOfAction(strHistogram);

    connect(MMath2->actions()[idMean],
            SIGNAL(triggered(bool)), this, SLOT(on_MeanValue_triggered(bool)));
    connect(MMath2->actions()[idMedian],
            SIGNAL(triggered(bool)), this, SLOT(on_MedianValue_triggered(bool)));
    connect(MMath2->actions()[idGauss],
            SIGNAL(triggered(bool)), this, SLOT(on_Gaussian_triggered(bool)));
    connect(MMath2->actions()[idHistogram],
            SIGNAL(toggled(bool)),   this, SLOT(on_Histogram_triggered(bool)));

    connect(AGXFitOrders, SIGNAL(triggered(QAction*)), this, SLOT(on_FitXOrder_changed()));
    connect(AGYFitOrders, SIGNAL(triggered(QAction*)), this, SLOT(on_FitYOrder_changed()));

    connect(AGSubtr, SIGNAL(triggered(QAction*)), this, SLOT(on_SubtrOrder_changed()));
    
    #ifndef _ADROID
    filterSettingsMenu->applyFilterSettings();
    #endif
    
    overlay->addUpButton(BPointSelection);
    overlay->addUpButton(BMarkers);
    overlay->addUpButton(BFilters);
    overlay->addRightButton(BPalette);
    overlay->addRightButton(BAreaZzoom);
    overlay->addRightButton(BSaveImage);
}

void ovControl_3d::InitFilters()
{
    BFilters = new TwoStateButton(":/ScanPlot/filterPlot2D.png", ":/ScanPlot/filterPlot2D.png");
    BFilters->setToolTip(QObject::tr("Enable filters"));
    BFilters->setFlat(true);
    overlay->setButtonStyle(BFilters);
    BFilters->hide();
    connect(BFilters, SIGNAL(state_changed(bool)), this, SLOT(on_Filter_changed(bool)));

    #ifndef _ADROID
    BFiltersSettings = new QPushButton();
    BFiltersSettings->setFlat(true);
    BFiltersSettings->setIcon(makeHugeIcon(":/MainToolbar/tools.png"));
    BFiltersSettings->setToolTip(QObject::tr("Filter settings"));
    BFiltersSettings->setParent(BFilters);
    {
        QString style = "";

        style += "QPushButton {background-color: rgba(255,255,255,0);";
        style += "border: 0px solid;}";
        style += "QToolTip { color: black; background-color: #FFFFE1; border: 1px solid black; }";
        style += "QPushButton:pressed{ background-color: lightgray; }";

        BFiltersSettings->setStyleSheet(style);
    }
    connect(BFiltersSettings, SIGNAL(released()), this, SLOT(filterSettingsClicked()));
#endif
}

void ovControl_3d::InitOldFilters()
{
    BOldFilter = new QPushButton();
    BOldFilter->hide();
    BOldFilter->setFlat(true);
    overlay->setButtonStyle(BOldFilter);
    BOldFilter->setIcon(makeHugeIcon(":/plot/math.png"));
    BOldFilter->setToolTip(QObject::tr("Filters"));

    BOldFilterAddit = new QPushButton();
    BOldFilterAddit->hide();
    BOldFilterAddit->setFlat(true);
    overlay->setButtonStyle(BOldFilterAddit);
    BOldFilterAddit->setIcon(makeHugeIcon(":/plot/math_extra.png"));
    BOldFilterAddit->setToolTip(QObject::tr("Addition math"));
}

void ovControl_3d::CreateActionsAndMenus()
{
    QMenu *MXFitOrders = new QMenu(tr("X Fitting"));
    QMenu *MYFitOrders = new QMenu(tr("Y Fitting"));
    QMenu *MSubtr = new QMenu(tr("Subtract"));

    AGXFitOrders = new QActionGroup(this);
    AGYFitOrders = new QActionGroup(this);
    AGSubtr      = new QActionGroup(this);

    const QString FitOrders[] = {tr("None"), tr("0 order"), tr("1st order"), tr("2nd order"), tr("3rd order"), tr("4th order"), tr("5th order")};
    for (auto &fitOrder:FitOrders)
    {
        AGXFitOrders->addAction(MXFitOrders->addAction(fitOrder))->setCheckable(true);
        AGYFitOrders->addAction(MYFitOrders->addAction(fitOrder))->setCheckable(true);
        AGSubtr->addAction(MSubtr->addAction(fitOrder))->setCheckable(true);
    }
    AGXFitOrders->actions()[0]->setChecked(true); //CurrentXFitOrder = -1;
    AGYFitOrders->actions()[0]->setChecked(true); //CurrentYFitOrder = -1;
    AGSubtr     ->actions()[0]->setChecked(true); //CurrentSubtrOrder = -1;

    //=================================================QMenu's=================================================
    MMathProcessing = new QMenu;
    MMathProcessing->addAction(tr("Fit Lines X"))->setCheckable(true);
    MMathProcessing->addAction(tr("Fit Lines Y"))->setCheckable(true);
    MMathProcessing->addMenu(MSubtr);
    BOldFilter->setMenu(MMathProcessing);

    MMath2 = new QMenu;
    MMath2->addMenu(MXFitOrders);
    MMath2->addMenu(MYFitOrders);
    MMath2->addSeparator();
    MMath2->addAction(T3dPlotControl::NameOfMeanValueFilter())->setCheckable(true);
    MMath2->addAction(T3dPlotControl::NameOfMedianValueFilter())->setCheckable(true);
    MMath2->addAction(T3dPlotControl::NameOfGaussianFilter())->setCheckable(true);
    MMath2->addAction(T3dPlotControl::NameOfHistogramFilter())->setCheckable(true);
    BOldFilterAddit->setMenu(MMath2);
}

void ovControl_3d::on_GwyXFit_triggered(bool Checked)
{
    parentPlot->Child->SetActivePlot(parentPlot->N);
    parentPlot->MakeCorrectLineByMedianaXDir(Checked);
    PlotControl->MMathProcessing->actions()[0]->setChecked(Checked);
    if (attributes)
        attributes->insert("Fit_lines_x", PlotControl->MMathProcessing->actions()[0]->isChecked());
}
void ovControl_3d::on_GwyYFit_triggered(bool Checked)
{
    parentPlot->Child->SetActivePlot(parentPlot->N);
    parentPlot->MakeCorrectLineByMedianaYDir(Checked);
    PlotControl->MMathProcessing->actions()[1]->setChecked(Checked);
    if (attributes)
        attributes->insert("Fit_lines_y", PlotControl->MMathProcessing->actions()[1]->isChecked());
}

void ovControl_3d::on_MeanValue_triggered(bool Checked)
{
    parentPlot->Child->SetActivePlot(parentPlot->N);
    auto name = T3dPlotControl::NameOfMeanValueFilter();
    int id = T3dPlotControl::GetNumOfAction(name);
    auto plotAction = PlotControl->MMath2->actions()[id];
    auto overAction = MMath2->actions()[id];

    parentPlot->MeanValueFilterEnabled = Checked;
    #ifndef _ADROID
    plotAction->setChecked(!Checked);
    #endif
    parentPlot->Child->on_MeanValue_triggered(Checked);
    Checked = plotAction->isChecked();
    parentPlot->MeanValueFilterEnabled = Checked;

    T3dPlot::SetFilterSizeTextInAction(overAction,
                              parentPlot->MeanValueFilterSize, name);
    if (parentPlot->MeanValueFilterSize == 0)
        overAction->setChecked(false);
    if (attributes)
        attributes->insert("mean_filter", parentPlot->MeanValueFilterSize);
}
void ovControl_3d::on_MedianValue_triggered(bool Checked)
{
    parentPlot->Child->SetActivePlot(parentPlot->N);
    auto name = T3dPlotControl::NameOfMedianValueFilter();
    int id = T3dPlotControl::GetNumOfAction(name);
    auto plotAction = PlotControl->MMath2->actions()[id];
    auto overAction = MMath2->actions()[id];

    parentPlot->MedianValueFilterEnabled = Checked;
    #ifndef _ADROID
    plotAction->setChecked(!Checked);
    #endif
    parentPlot->Child->on_MedianValue_triggered(Checked);
    Checked = plotAction->isChecked();
    parentPlot->MedianValueFilterEnabled = Checked;

    T3dPlot::SetFilterSizeTextInAction(overAction,
                              parentPlot->MedianValueFilterSize, name);
    if (parentPlot->MedianValueFilterSize == 0)
        overAction->setChecked(false);
    if (attributes)
        attributes->insert("median_filter", parentPlot->MedianValueFilterSize);
}
void ovControl_3d::on_Gaussian_triggered(bool Checked)
{
    parentPlot->Child->SetActivePlot(parentPlot->N);
    auto name = T3dPlotControl::NameOfGaussianFilter();
    int id = T3dPlotControl::GetNumOfAction(name);
    auto plotAction = PlotControl->MMath2->actions()[id];
    auto overAction = MMath2->actions()[id];

    parentPlot->GaussianValueFilterEnabled = Checked;
    #ifndef _ADROID
    plotAction->setChecked(!Checked);
    #endif
    parentPlot->Child->on_Gaussian_triggered(Checked);
    Checked = plotAction->isChecked();
    parentPlot->GaussianValueFilterEnabled = Checked;

    T3dPlot::SetFilterSizeTextInAction(overAction,
                              parentPlot->GaussianFilterSize, name);
    if (parentPlot->GaussianFilterSize == 0)
        overAction->setChecked(false);
    if (attributes)
        attributes->insert("gaussian_filter", parentPlot->GaussianFilterSize);
}
void ovControl_3d::on_Histogram_triggered(bool Checked)
{
    auto name = T3dPlotControl::NameOfHistogramFilter();
    int id = T3dPlotControl::GetNumOfAction(name);

    PlotControl->MMath2->actions()[id]->trigger();
    PlotControl->MMath2->actions()[id]->setChecked(Checked);

    T3dPlot::SetFilterSizeTextInAction(MMath2->actions()[id], 0, name);
    if (attributes)
        attributes->insert("histogram_filter", Checked);
}

void ovControl_3d::on_FitXOrder_changed()
{
    int id = FindCheckedItemIn(AGXFitOrders);
    PlotControl->AGXFitOrders->actions()[id]->setChecked(true);
    PlotControl->AGXFitOrders->actions()[id]->trigger();
    if (attributes)
        attributes->insert("x_fitting", id);
}
void ovControl_3d::on_FitYOrder_changed()
{
    int id = FindCheckedItemIn(AGYFitOrders);
    PlotControl->AGYFitOrders->actions()[id]->setChecked(true);
    PlotControl->AGYFitOrders->actions()[id]->trigger();
    if (attributes)
        attributes->insert("y_fitting", id);
}
void ovControl_3d::on_SubtrOrder_changed()
{
    int id = FindCheckedItemIn(AGSubtr);
    PlotControl->AGSubtr->actions()[id]->trigger();
    if (attributes)
        attributes->insert("substract_order", id);
}

void ovControl_3d::onPointSelection_changed()
{
    parentPlot->Child->SetActivePlot(parentPlot->N);
    PlotControl->BCross->change_state_man_with_sig(true);
    overlay->setBackButtonVisible(true);
    overlay->setVisibilityMode(plot::ThreeDOverlayButtons::svm_point_selection);
    #ifndef _ADROID
    if (parentPlot->Child->TwoDPlotVertical->isHidden() && parentPlot->Child->ThreeDPlots.count() == 1)
        parentPlot->Child->SetFullWindowPlot(0, true);
    #endif
    overlay->setVisibleIfUnderMouse();
    overlay->blockResizeTimes = 0;
}

void ovControl_3d::onMarkers_changed()
{
    parentPlot->Child->SetActivePlot(parentPlot->N);
    PlotControl->BMarkers->setChecked(true);
    overlay->setVisibilityMode(plot::ThreeDOverlayButtons::svm_markers);
    overlay->setBackButtonVisible(true);
    #ifndef _ADROID
    if (parentPlot->Child->TwoDPlotVertical->isHidden() && parentPlot->Child->ThreeDPlots.count() == 1)
        parentPlot->Child->SetFullWindowPlot(0, true);
    #endif
    overlay->setVisibleIfUnderMouse();
    overlay->blockResizeTimes = 0;
}

void ovControl_3d::on_Filter_changed(bool pressed)
{
    parentPlot->Child->SetActivePlot(parentPlot->N);
    overlay->setButtonActive(BFilters, pressed);
    parentPlot->setFiltering(pressed);
    #ifndef _ADROID
    auto filterSetSize = overlay->UpButtonSize/2;
    BFiltersSettings->setFixedSize(filterSetSize, filterSetSize);
    BFiltersSettings->setIconSize(QSize(filterSetSize, filterSetSize));
    #endif
}

void ovControl_3d::filterSettingsClicked()
{
    parentPlot->Child->SetActivePlot(parentPlot->N);
    overlay->setVisibilityMode(plot::ThreeDOverlayButtons::svm_filter_settings);
    overlay->setBackButtonVisible(true);
    overlay->setVisibleMenu(true);
}

void ovControl_3d::colorPickClicked()
{
    colorPickClicked(overlay->visibilityMode);
}

void ovControl_3d::colorPickClicked(plot::ThreeDOverlayButtons::SpecialVisibilityMode visibilityMode)
{
    parentPlot->Child->SetActivePlot(parentPlot->N);
    
    if (visibilityMode == plot::ThreeDOverlayButtons::svm_palettes)
    {
        overlay->setVisibilityMode(plot::ThreeDOverlayButtons::svm_none);
        overlay->setBackButtonVisible(false);
    }
    else
    {
        overlay->setVisibilityMode(plot::ThreeDOverlayButtons::svm_palettes);
        overlay->setBackButtonVisible(true);
    }
    overlay->setVisibleIfUnderMouse();
}

void ovControl_3d::on3DBAreaZoom_changed()
{
    parentPlot->Child->SetActivePlot(parentPlot->N);
    parentPlot->AreaZzoom = true;
    overlay->setVisibilityMode(plot::ThreeDOverlayButtons::svm_range_selection);
    overlay->setBackButtonVisible(true);
    overlay->setVisibleIfUnderMouse();
}

void ovControl_3d::onSaveImage_clicked()
{
    parentPlot->Child->SetActivePlot(parentPlot->N);
    overlay->blockShowing3DPlotButtons = true;
    overlay->setVisibleMenu(false);
    parentPlot->Child->on_BSavePlotPixmap_clicked();
    overlay->blockShowing3DPlotButtons = false;
    overlay->setVisibleIfUnderMouse();
}

void ovControl_3d::setVisibleFilterButton(bool visible)
{
    if (hideFilterButton == !visible)
        return;

    hideFilterButton = !visible;
    if (visible)
    {
        overlay->addUpButton(BFilters, 3);
    }
    else {
        overlay->removeButton(BFilters);
    }
    overlay->updateResize();
}

void ovControl_3d::ActivateRatisMode()
{
    overlay->removeButton(BPalette);
    overlay->removeButton(BMarkers);
    overlay->removeButton(BSaveImage);
    overlay->removeButton(BFilters);
    overlay->removeButton(BAreaZzoom);
}

void ovControl_3d::resetOldFilters()
{
    attributes = nullptr;
    parentPlot->isFiltering = true;
    parentPlot->MeanValueFilterSize = 0;
    parentPlot->MedianValueFilterSize = 0;
    parentPlot->GaussianFilterSize = 0;
    AGXFitOrders->actions()[0]->setChecked(true);
    AGYFitOrders->actions()[0]->setChecked(true);
    AGSubtr     ->actions()[0]->setChecked(true);
    MMathProcessing->actions().at(0)->setChecked(false);
    MMathProcessing->actions().at(1)->setChecked(false);

    QString strMean = T3dPlotControl::NameOfMeanValueFilter();
    QString strMedian = T3dPlotControl::NameOfMedianValueFilter();
    QString strGauss = T3dPlotControl::NameOfGaussianFilter();
    QString strHistogram = T3dPlotControl::NameOfHistogramFilter();
    int idMean = T3dPlotControl::GetNumOfAction(strMean);
    int idMedian = T3dPlotControl::GetNumOfAction(strMedian);
    int idGauss = T3dPlotControl::GetNumOfAction(strGauss);
    int idHistogram = T3dPlotControl::GetNumOfAction(strHistogram);
    MMath2->actions()[idMean]->setChecked(false);
    MMath2->actions()[idMedian]->setChecked(false);
    MMath2->actions()[idGauss]->setChecked(false);
    MMath2->actions()[idHistogram]->setChecked(false);
    MMath2->actions()[idMean]->setText(strMean);
    MMath2->actions()[idMedian]->setText(strMedian);
    MMath2->actions()[idGauss]->setText(strGauss);
    parentPlot->Child->SetAllFilterSizesTextInAction();
}

void ovControl_3d::setVisibleOldFilters(bool visible)
{
    isOldFiltersVisible = visible;
    if (visible)
    {
        overlay->addUpButton(BOldFilter);
        overlay->addUpButton(BOldFilterAddit);

        resetOldFilters();
    }
    else
    {
        BOldFilter->hide();
        BOldFilterAddit->hide();
        overlay->removeButton(BOldFilter);
        overlay->removeButton(BOldFilterAddit);
    }
    overlay->updateResize();
}

void ovControl_3d::setVisiblePointSelection(bool visible)
{
    if (visible)
        overlay->addUpButton(BPointSelection, 0);
    else
    {
        BPointSelection->hide();
        overlay->removeButton(BPointSelection);
    }
    overlay->updateResize();
}

void ovControl_3d::setAttributes(plot::TMetaStorage *attr)
{
    attributes = attr;

    if (attr == nullptr)
        return;

    if (attr->contains("Fit_lines_x"))
        MMathProcessing->actions()[0]->setChecked(attr->value("Fit_lines_x").toBool());
    if (attr->contains("Fit_lines_y"))
        MMathProcessing->actions()[1]->setChecked(attr->value("Fit_lines_y").toBool());
    if (attr->contains("substract_order"))
    {
        AGSubtr->actions()[attr->value("substract_order").toInt()]->trigger();
        AGSubtr->actions()[attr->value("substract_order").toInt()]->setChecked(true);
    }
    if (attr->contains("x_fitting"))
    {
        AGXFitOrders->actions()[attr->value("x_fitting").toInt()]->trigger();
        AGXFitOrders->actions()[attr->value("x_fitting").toInt()]->setChecked(true);
    }
    if (attr->contains("y_fitting"))
    {
        AGYFitOrders->actions()[attr->value("y_fitting").toInt()]->trigger();
        AGYFitOrders->actions()[attr->value("y_fitting").toInt()]->setChecked(true);
    }
    #ifndef _ADROID
    if (attr->contains("mean_filter"))
    {
        int value = attr->value("mean_filter").toInt();

        if (value > 0)
        {
            auto name = T3dPlotControl::NameOfMeanValueFilter();
            int id = T3dPlotControl::GetNumOfAction(name);
            parentPlot->MeanValueFilterEnabled = true;
            parentPlot->MeanValueFilterSize = value;
            PlotControl->MMath2->actions()[id]->setChecked(true);
            MMath2->actions()[id]->setChecked(true);

            T3dPlot::SetFilterSizeTextInAction(PlotControl->MMath2->actions()[id],
                                               parentPlot->MeanValueFilterSize, name);
            T3dPlot::SetFilterSizeTextInAction(MMath2->actions()[id],
                                               parentPlot->MeanValueFilterSize, name);
        }
    }
    if (attr->contains("median_filter"))
    {
        int value = attr->value("median_filter").toInt();

        if (value > 0)
        {
            auto name = T3dPlotControl::NameOfMedianValueFilter();
            int id = T3dPlotControl::GetNumOfAction(name);
            parentPlot->MedianValueFilterEnabled = true;
            parentPlot->MedianValueFilterSize = value;
            PlotControl->MMath2->actions()[id]->setChecked(true);
            MMath2->actions()[id]->setChecked(true);

            T3dPlot::SetFilterSizeTextInAction(PlotControl->MMath2->actions()[id],
                                               parentPlot->MedianValueFilterSize, name);
            T3dPlot::SetFilterSizeTextInAction(MMath2->actions()[id],
                                               parentPlot->MedianValueFilterSize, name);
        }
    }
    if (attr->contains("gaussian_filter"))
    {
        int value = attr->value("gaussian_filter").toInt();

        if (value > 0)
        {
            auto name = T3dPlotControl::NameOfGaussianFilter();
            int id = T3dPlotControl::GetNumOfAction(name);
            parentPlot->GaussianValueFilterEnabled = true;
            parentPlot->GaussianFilterSize = value;
            PlotControl->MMath2->actions()[id]->setChecked(true);
            MMath2->actions()[id]->setChecked(true);

            T3dPlot::SetFilterSizeTextInAction(PlotControl->MMath2->actions()[id],
                                               parentPlot->GaussianFilterSize, name);
            T3dPlot::SetFilterSizeTextInAction(MMath2->actions()[id],
                                               parentPlot->GaussianFilterSize, name);
        }
    }
    if (attr->contains("histogram_filter"))
    {
        auto histogram = MMath2->actions()[T3dPlotControl::GetNumOfAction(T3dPlotControl::NameOfHistogramFilter())];
        histogram->trigger();
        histogram->setChecked(attr->value("histogram_filter").toBool());
    }
    #endif
}

void ovControl_3d::onHelpBtn_clicked()
{
    parentPlot->Child->SetActivePlot(parentPlot->N);
    overlay->setVisibilityMode(plot::ThreeDOverlayButtons::svm_help);
    overlay->setBackButtonVisible(true);
    overlay->setVisibleMenu(true);
    parentPlot->blockWheelZooming = true;
}

void ovControl_3d::setHelpImage(const QString path)
{
    QPixmap img = QPixmap(path);
    auto lab = new QLabel();
    lab->setPixmap(img);
    helpScrollArea->setWidget(lab);
    helpImgWidth = img.width() + 17;
    helpSetted = true;
    //allWidgets.append(helpScrollArea);
}

void ovControl_3d::ApplyEnterClicked(plot::ThreeDOverlayButtons::SpecialVisibilityMode visibilityMode)
{
    if (visibilityMode == plot::ThreeDOverlayButtons::svm_area_selection)
    {
        parentPlot->Child->on_BApply_clicked();
        for (auto p:parentPlot->Child->ThreeDPlots)
        {
            p->AreaZzoom = false;
            p->overlayButtons->setVisibilityMode(plot::ThreeDOverlayButtons::svm_none);
        }
    }
}

void ovControl_3d::BackCancelClicked(plot::ThreeDOverlayButtons::SpecialVisibilityMode visibilityMode)
{
    parentPlot->Child->SetActivePlot(parentPlot->N);
    
    if (visibilityMode == plot::ThreeDOverlayButtons::svm_area_selection)
    {
        parentPlot->Child->on_BCancel_clicked();
        for (auto p:parentPlot->Child->ThreeDPlots)
        {
            p->AreaZzoom = false;
            p->overlayButtons->setVisibilityMode(plot::ThreeDOverlayButtons::svm_none);
        }
    }
    else if (visibilityMode == plot::ThreeDOverlayButtons::svm_point_selection)
    {
        PlotControl->BCross->change_state_man_with_sig(false);
    }
    else if (visibilityMode == plot::ThreeDOverlayButtons::svm_markers)
    {
        PlotControl->BMarkers->setChecked(false);
    }
    else if (visibilityMode == plot::ThreeDOverlayButtons::svm_range_selection)
    {
        parentPlot->AreaZzoom = false;
    }
    else if (visibilityMode == plot::ThreeDOverlayButtons::svm_palettes)
        colorPickClicked(visibilityMode);
    else if (visibilityMode == plot::ThreeDOverlayButtons::svm_help)
        parentPlot->blockWheelZooming = false;
}

void ovControl_3d::setVisiblePlot(bool visible, plot::ThreeDOverlayButtons::SpecialVisibilityMode visibilityMode)
{
    bool visibleHelp = visible && (visibilityMode == plot::ThreeDOverlayButtons::svm_help);
    helpScrollArea->setVisible(visibleHelp);
    
    overlay->setButtonVisible(BFilters, !hideFilterButton && visible && (visibilityMode == plot::ThreeDOverlayButtons::svm_none));
    
    overlay->setButtonVisible(helpButton, visible && showHelpButton && helpSetted && (visibilityMode == plot::ThreeDOverlayButtons::svm_none));
    helpButton->setIconSize(QSize(helpButton->width(), helpButton->height()));
}

void ovControl_3d::resized(QSize newSize)
{
    #ifndef _ADROID
    auto filterSetSize = overlay->UpButtonSize/2;
    BFiltersSettings->setFixedSize(filterSetSize, filterSetSize);
    BFiltersSettings->setIconSize(QSize(filterSetSize, filterSetSize));
    BFiltersSettings->move(filterSetSize, 0);
#endif
    
    int y = overlay->BBackButton->pos().y() + overlay->BBackButton->height();
    int helpScrollAreaHeight = overlay->buttonRect.height() - overlay->BBackButton->height();
    
    helpScrollArea->move(overlay->buttonRect.x(), y);
    int helpScrollAreaWidth = newSize.width() - overlay->buttonRect.x() - overlay->btnsDistBetw;
    helpScrollAreaWidth = qMin(helpScrollAreaWidth, helpImgWidth);
    helpScrollArea->setFixedSize(helpScrollAreaWidth, helpScrollAreaHeight);
}

}
