#include "ovmenu_3d_filtersettings.h"

#include <QObject>
#include <QWidget>
#include "3d/plot_3dplot.h"
#include "3d/plot_3dpure.h"
#include "../../src/gui/controls.h"
#include "qwt_plot_layout.h"
#include "core/plot_core.h"
#include <QAbstractItemView>
#include <QTimer>
#include "3d/ovcontrol_3d.h"

namespace plot {

ovMenu_3d_FilterSettings::ovMenu_3d_FilterSettings(ThreeDOverlayButtons *overlay) : colorRect2(QColor(240, 240, 240))
{
    this->overlay = overlay;
    this->setParent(overlay->parentPlot);
    this->hide();
    connect(overlay, &ThreeDOverlayButtons::GUIPlotSetVisible, this, &ovMenu_3d_FilterSettings::setVisiblePlot);
    connect(overlay, &ThreeDOverlayButtons::GUIPlotResized, this, &ovMenu_3d_FilterSettings::resized);
    
    chFitLines = new QCheckBox();
    cbSubstract = new QComboBox();
    cbFitting = new QComboBox();
    chMeanValue = new QCheckBox();
    chMedianValue = new QCheckBox();
    chGaussian = new QCheckBox();
    chHistogramEqualization = new QCheckBox();
    edMeanValue = new Edit(2, 30);
    edMedianValue = new Edit(2, 30);
    edGaussian = new Edit(2, 30);

    widMean = new QWidget();
    widMedian = new QWidget();
    widGaussian = new QWidget();

    widMean->setContentsMargins(0, 0, 0, 0);
    widMedian->setContentsMargins(0, 0, 0, 0);
    widGaussian->setContentsMargins(0, 0, 0, 0);

    chGaussian->setSizePolicy(QSizePolicy::Fixed, QSizePolicy::Fixed);
    chMeanValue->setSizePolicy(QSizePolicy::Fixed, QSizePolicy::Fixed);
    chMedianValue->setSizePolicy(QSizePolicy::Fixed, QSizePolicy::Fixed);

    auto layMean = new QHBoxLayout(widMean);
    auto layMedian = new QHBoxLayout(widMedian);
    auto layGaussian = new QHBoxLayout(widGaussian);

    layMean->setMargin(0);
    layMedian->setMargin(0);
    layGaussian->setMargin(0);

    layMean->addWidget(chMeanValue, 0, Qt::AlignRight);
    layMean->addWidget(edMeanValue, 0, Qt::AlignRight);

    layMedian->addWidget(chMedianValue, 0, Qt::AlignRight);
    layMedian->addWidget(edMedianValue, 0, Qt::AlignRight);

    layGaussian->addWidget(chGaussian, 0, Qt::AlignRight);
    layGaussian->addWidget(edGaussian, 0, Qt::AlignRight);

    edMeanValue->setSizePolicy(QSizePolicy::Minimum, QSizePolicy::Fixed);
    edMedianValue->setSizePolicy(QSizePolicy::Minimum, QSizePolicy::Fixed);
    edGaussian->setSizePolicy(QSizePolicy::Minimum, QSizePolicy::Fixed);

    const QString FitOrders[] = {tr("None"), tr("0 ord"), tr("1st ord"), tr("2nd ord"), tr("3rd ord"), tr("4th ord"), tr("5th ord")};
    for (auto &fitOrder:FitOrders)
    {
        cbSubstract->addItem(fitOrder);
        cbFitting->addItem(fitOrder);
    }

    auto labFitLines = new QLabel(tr("Fit lines"));
    auto labSubstract = new QLabel(tr("Substract"));
    auto labFitting = new QLabel(tr("Fitting"));
    auto labMeanValue = new QLabel(tr("Mean value"));
    auto labMedianValue = new QLabel(tr("Median value"));
    auto labGaussian = new QLabel(tr("Gaussian"));
    auto labHistogramEqualization = new QLabel(tr("Hist. equals"));

    filterSettingsLabels << labFitLines << labSubstract << labFitting << labMeanValue
                        << labMedianValue << labGaussian << labHistogramEqualization;
    filterSettingsRightPart << chFitLines << cbSubstract << cbFitting << widMean
                            << widMedian << widGaussian << chHistogramEqualization;

    this->setStyleSheet("QSplitter {border: 40px solid black; border-style: outset; border-width: 40px; border-radius: 10px; border-color: #B4B4B4; background-color: #E3E3EC;}");

    QString style = "color: black; font: 17px;";
    for (auto lab:filterSettingsLabels)
    {
        lab->setStyleSheet(style);
        lab->setParent(this);
    }
    for (auto w:filterSettingsRightPart)
    {
        w->setParent(this);
    }
    
    loadFilterSettings();

    connect(chFitLines, SIGNAL(toggled(bool)), this, SLOT(saveScanFilterSettings()));
    connect(cbSubstract, SIGNAL(currentIndexChanged(int)), this, SLOT(saveScanFilterSettings()));
    connect(cbFitting, SIGNAL(currentIndexChanged(int)), this, SLOT(saveScanFilterSettings()));
    connect(chMeanValue, SIGNAL(toggled(bool)), this, SLOT(saveScanFilterSettings()));
    connect(chMedianValue, SIGNAL(toggled(bool)), this, SLOT(saveScanFilterSettings()));
    connect(chGaussian, SIGNAL(toggled(bool)), this, SLOT(saveScanFilterSettings()));
    connect(chHistogramEqualization, SIGNAL(toggled(bool)), this, SLOT(saveScanFilterSettings()));
    connect(edMeanValue, SIGNAL(newValue()), this, SLOT(saveScanFilterSettings()));
    connect(edMedianValue, SIGNAL(newValue()), this, SLOT(saveScanFilterSettings()));
    connect(edGaussian, SIGNAL(newValue()), this, SLOT(saveScanFilterSettings()));
    //---------
}

void ovMenu_3d_FilterSettings::loadFilterSettings()
{
    QSettings settings("Param/scan.conf", QSettings::IniFormat);
    settings.beginGroup("3D_plot_filter");
    chFitLines->setChecked(settings.value("fit_lines", false).toBool());
    cbSubstract->setCurrentIndex(settings.value("substract", 0).toInt());
    cbFitting->setCurrentIndex(settings.value("fitting", 0).toInt());
    chMeanValue->setChecked(settings.value("mean_value", false).toBool());
    chMedianValue->setChecked(settings.value("median_value", false).toBool());
    chGaussian->setChecked(settings.value("gaussian", false).toBool());
    chHistogramEqualization->setChecked(settings.value("histogam_equalization", false).toBool());
    edMeanValue->setValue(settings.value("ed_mean_value", 2).toInt());
    edMedianValue->setValue(settings.value("ed_median_value", 2).toInt());
    edGaussian->setValue(settings.value("ed_gaussian", 2).toInt());
    settings.endGroup();
}

void ovMenu_3d_FilterSettings::setVisiblePlot(bool visible, ThreeDOverlayButtons::SpecialVisibilityMode visibilityMode)
{
    bool visibleFilterSettingsMenu = visible && visibilityMode == ThreeDOverlayButtons::SpecialVisibilityMode::svm_filter_settings;
    this->setVisible(visibleFilterSettingsMenu);
}

void ovMenu_3d_FilterSettings::resized(QSize newSize)
{
    const int edge = 20;
    const int distanceBetween = 10;
    int maxRightX = 0;
    int cbsHeight = 0;
    int maxLeftWidth = 0;
    int maxRightWidth = 0;
    for (int i=0; i<filterSettingsLabels.size(); i++)
    {
        int widthLab = filterSettingsLabels.at(i)->minimumSizeHint().width();
        if (widthLab > maxLeftWidth)
            maxLeftWidth = widthLab;
        int widthRight = filterSettingsRightPart.at(i)->minimumSizeHint().width();
        if (widthRight > maxRightWidth)
            maxRightWidth = widthRight;
        cbsHeight += filterSettingsRightPart.at(i)->minimumSizeHint().height();
        cbsHeight += distanceBetween;
    }
    maxRightX = maxLeftWidth + edge + maxRightWidth;
    
    cbSubstract->setFixedWidth(maxRightWidth);
    cbFitting->setFixedWidth(maxRightWidth);
    widGaussian->setFixedWidth(maxRightWidth);
    edGaussian->setFixedWidth(maxRightWidth - chGaussian->minimumSizeHint().width() - 26);
    widMean->setFixedWidth(maxRightWidth);
    edMeanValue->setFixedWidth(maxRightWidth - chMeanValue->minimumSizeHint().width() - 26);
    widMedian->setFixedWidth(maxRightWidth);
    edMedianValue->setFixedWidth(maxRightWidth - chMedianValue->minimumSizeHint().width() - 26);

    int backWidth = maxRightX + edge*1.5;
    int backHeight = cbsHeight + edge*1.5;
    this->setFixedSize(backWidth, backHeight);
    int x = newSize.width()*0.5 - backWidth*0.5;
    int y = newSize.height()*0.5 - backHeight*0.5;
    this->move(x, y);
    x = 0;
    y = 0;
    x += edge;
    y += edge;
    for (int i=0; i<filterSettingsLabels.size(); i++)
    {
        int rWidth = qMin(maxRightWidth, filterSettingsRightPart.at(i)->width());
        
        filterSettingsLabels.at(i)->move(x, y + filterSettingsLabels.at(i)->minimumSizeHint().height()*0.2);
        filterSettingsRightPart.at(i)->move(x + maxRightX - rWidth, y);
        y += (filterSettingsLabels.at(i)->minimumSizeHint().height() < filterSettingsRightPart.at(i)->minimumSizeHint().height() ?
                  filterSettingsRightPart.at(i)->minimumSizeHint().height():filterSettingsLabels.at(i)->minimumSizeHint().height());
        y += distanceBetween;
    }
}

void ovMenu_3d_FilterSettings::saveScanFilterSettings()
{
    QSettings settings("Param/scan.conf", QSettings::IniFormat);
    settings.beginGroup("3D_plot_filter");
    settings.setValue("fit_lines", chFitLines->isChecked());
    settings.setValue("substract", cbSubstract->currentIndex());
    settings.setValue("fitting", cbFitting->currentIndex());
    settings.setValue("mean_value", chMeanValue->isChecked());
    settings.setValue("median_value", chMedianValue->isChecked());
    settings.setValue("gaussian", chGaussian->isChecked());
    settings.setValue("histogam_equalization", chHistogramEqualization->isChecked());
    settings.setValue("ed_mean_value", edMeanValue->value());
    settings.setValue("ed_median_value", edMedianValue->value());
    settings.setValue("ed_gaussian", edGaussian->value());
    settings.endGroup();

    applyFilterSettings();
    
    for (auto plot:overlay->parentPlot->Child->ThreeDPlots)
    {
        if (plot->N != overlay->parentPlot->N)
            plot->overlayButtons->control->filterSettingsMenu->loadFilterSettings();
    }
}

void ovMenu_3d_FilterSettings::applyFilterSettings()
{
    bool xDirScan = overlay->parentPlot->Child->ThreeDPlots.first()->xScanning;
    
    if (overlay->PlotControl)
    {
        overlay->PlotControl->MMathProcessing->actions()[0]->setChecked((xDirScan ? chFitLines->isChecked():false));
        //PlotControl->MMathProcessing->actions()[0]->trigger();
        overlay->PlotControl->MMathProcessing->actions()[1]->setChecked((xDirScan ? false:chFitLines->isChecked()));
        //PlotControl->MMathProcessing->actions()[1]->trigger();
        overlay->PlotControl->AGSubtr->actions()[cbSubstract->currentIndex()]->trigger();
        overlay->PlotControl->AGXFitOrders->actions()[xDirScan ? cbFitting->currentIndex():0]->setChecked(true);
        //PlotControl->AGXFitOrders->actions()[xDirScan ? cbFitting->currentIndex():0]->trigger();
        overlay->PlotControl->AGYFitOrders->actions()[xDirScan ? 0:cbFitting->currentIndex()]->setChecked(true);
        //PlotControl->AGYFitOrders->actions()[xDirScan ? 0:cbFitting->currentIndex()]->trigger();
        auto mmathActions = overlay->PlotControl->MMath2->actions();
        mmathActions[overlay->PlotControl->GetNumOfAction(overlay->PlotControl->NameOfMeanValueFilter())]->
            setChecked(chMeanValue->isChecked());
        mmathActions[overlay->PlotControl->GetNumOfAction(overlay->PlotControl->NameOfMedianValueFilter())]->
            setChecked(chMedianValue->isChecked());
        mmathActions[overlay->PlotControl->GetNumOfAction(overlay->PlotControl->NameOfGaussianFilter())]->
            setChecked(chGaussian->isChecked());
        mmathActions[overlay->PlotControl->GetNumOfAction(overlay->PlotControl->NameOfHistogramFilter())]->
            setChecked(chHistogramEqualization->isChecked());
    }
    
    auto plot = overlay->parentPlot;
    plot->MakeCorrectLineByMedianaXDir(xDirScan ? chFitLines->isChecked():false);
    plot->MakeCorrectLineByMedianaYDir(xDirScan ? false:chFitLines->isChecked());
    plot->MakeSubtr(cbSubstract->currentIndex() - 1);
    plot->MakeXFit((xDirScan ? cbFitting->currentIndex():0) - 1);
    plot->MakeYFit((xDirScan ? 0:cbFitting->currentIndex()) - 1);
    plot->MeanValueFilterEnabled = chMeanValue->isChecked();
    plot->MedianValueFilterEnabled = chMedianValue->isChecked();
    plot->GaussianValueFilterEnabled = chGaussian->isChecked();
    plot->MeanValueFilterSize = edMeanValue->value();
    plot->MedianValueFilterSize = edMedianValue->value();
    plot->GaussianFilterSize = edGaussian->value();
    plot->HistogramEqualization = chHistogramEqualization->isChecked();
    plot->ShowHistogram(chHistogramEqualization->isChecked());
    plot->Refresh();
    
    overlay->parentPlot->Child->SetAllFilterSizesTextInAction();
}

}
