#include "ars3000.h"
#include <QSettings>
#include "motors.hpp"
#include <QtDebug>
#include <QComboBox>

double ARS3000::ars_slit_pos[2] = {-1, -1};
double ARS3000::ars_laser_pos = -1;

ARS3000::ARS3000() : QGroupBox()
{
    for (int i=0; i<numMotorRoles; i++)
        motorCurrentStages[i] = -1;
    
    loadSettings();
    initGUI();
    
    QSettings settings("Param/ARS3000.conf", QSettings::IniFormat);
    auto step = settings_double(settings, "ARS3000_objslmonohr_step", 1);
    auto speed = settings_double(settings, "ARS3000_objslmonohr_speed", 1);
    objsl_stepValue[0]->setValue(step);
    objsl_speedValue[0]->setValue(speed);
    step = settings_double(settings, "ARS3000_objslpmt_step", 1);
    speed = settings_double(settings, "ARS3000_objslpmt_speed", 1);
    objsl_stepValue[1]->setValue(step);
    objsl_speedValue[1]->setValue(speed);
}

void ARS3000::setMotors(Motors *m)
{
    connect(m, SIGNAL(motorStatusChanged(int, bool, bool)), this, SLOT(motorEvent(int, bool, bool)));
    connect(m, SIGNAL(motorSendingStatusEnded()), this, SLOT(updateQueue()));
}

void ARS3000::setSelectedConfLine(int motor, int confLine)
{
    ARS3000_selectedConfLines[motor] = confLine;
    
    auto m = Motors::GetIt();
    auto role = m->mapVirtToRole(motor+1);
    emit confLineChanged(role, confLine);
}

void ARS3000::setActiveMotor(int num, Motors::MotorRoles role)
{
    this->setTitle(title + ", " + Motors::axes_names[role]);
    
    currentMotorNum = num;
    currentMotorRole = role;
    
    objsl_wid[0]->setVisible(role == Motors::ObjXSlMonochr);
    objsl_wid[1]->setVisible(role == Motors::ObjXSlPMT);
    
    ARS3000_numConfLinesGui->blockSignals(true);
    ARS3000_numConfLinesGui->setValue(ARS3000_numConfLines[num]);
    ARS3000_numConfLinesGui->blockSignals(false);
    
    ARS3000_limitterGui->blockSignals(true);
    ARS3000_limitterGui->setChecked(ARS3000_limitter[num]);
    ARS3000_limitterGui->blockSignals(false);
    
    ARS3000_radioSelectedConfLine[ARS3000_selectedConfLines[num]]->blockSignals(true);
    ARS3000_radioSelectedConfLine[ARS3000_selectedConfLines[num]]->setChecked(true);
    ARS3000_radioSelectedConfLine[ARS3000_selectedConfLines[num]]->blockSignals(false);
    
    for (int i=0; i<ARS3000_maxNumConfLines; i++)
    {
        ARS3000_textboxConfLine[i]->blockSignals(true);
        ARS3000_textboxConfLine[i]->setText(ARS3000_confLinesText[num][i]);
        ARS3000_textboxConfLine[i]->blockSignals(false);
        ARS3000_confLinesValuesGui[i]->blockSignals(true);
        ARS3000_confLinesValuesGui[i]->setValue(ARS3000_confLinesValues[num][i]);
        ARS3000_confLinesValuesGui[i]->blockSignals(false);
    }
    
    numLinesChanged();
}

void ARS3000::addMotorToStartQueue(Motors::MotorRoles role)
{
    auto m = Motors::GetIt();
    auto motor = m->getRoleMotor(role);
    
    if (!m->isMotorExist(role))
        return;
    
    motorQueueStruct st;
    st.role = role;
    st.dist = ARS3000_confLinesValues[motor-1][ARS3000_selectedConfLines[motor-1]];
    motorQueue.enqueue(st);
    
    updateQueue();
}

void ARS3000::updateQueue()
{
    if (motorQueue.isEmpty())
        return;
    
    for (int i=0; i<numMotorRoles; i++)
    {
        if (motorCurrentStages[i] != -1)
            return; //stop queueing if any motor is active
    }
    
    auto st = motorQueue.dequeue();
    startMotor(true, st.role);
}

void ARS3000::startMotor(bool start, Motors::MotorRoles role)
{
    auto m = Motors::GetIt();
    
    if (!m->isMotorExist(role))
        return;
    
    auto motor = m->getRoleMotor(role);
    int roleIndex = role - Motors::NDFilter;
    if (start)
    {
        motorCurrentStages[roleIndex] = 0;
        int speed = m->getMotorMaxSpeed(motor);
        if (ARS3000_limitter[motor-1])
            m->goDownE(motor, speed);
        else
            m->goUpE(motor, speed);
    }
    else
    {
        motorCurrentStages[roleIndex] = -1;
        if (ARS3000_start->isOn)
            ARS3000_start->change_state_man(false);
        m->goStop(motor);
    }
}

void ARS3000::startMotorWithConfLine(Motors::MotorRoles role, int confLine)
{
    auto m = Motors::GetIt();
    
    if (!m->isMotorExist(role))
        return;
    
    auto motor = m->getRoleMotor(role);
    if (confLine < ARS3000_numConfLines[motor - 1])
    {
        setSelectedConfLine(motor - 1, confLine);
        startMotor(true, role);
    }
}

void ARS3000::stopAllMotors()
{
    motorQueue.clear();
    for (int i=0; i<numMotorRoles; i++)
        if (motorCurrentStages[i] != -1)
            startMotor(false, (Motors::MotorRoles)(Motors::NDFilter + i));
}

void ARS3000::motorEvent(int motor, bool atEnd, bool active)
{
    auto m = Motors::GetIt();
    Motors::MotorRoles role = m->mapVirtToRole(motor);
    
    if (!isMotorRoleARS3000(role))
        return;
    
    int roleIndex = role - Motors::NDFilter;
    if (motorCurrentStages[roleIndex] == 0)
    {
        if (active) return;
        
        motorCurrentStages[roleIndex] = 1;
        int dist = ARS3000_confLinesValues[motor-1][ARS3000_selectedConfLines[motor-1]];
        int speed = m->getMotorMaxSpeed(motor);
        if (ARS3000_limitter[motor-1])
            m->goUp(motor, speed, dist);
        else
            m->goDown(motor, speed, dist);
        return;
    }
    
    if (motorCurrentStages[roleIndex] == 1)
    {
        if (active) return;

        motorCurrentStages[roleIndex] = -1;
        if (ARS3000_start->isOn)
            ARS3000_start->change_state_man(false);
        emit motorStopped(motor, role);
    }
}

void ARS3000::start(bool start)
{
    startMotor(start, currentMotorRole);
}

void ARS3000::fillComboBoxConfLines(Motors::MotorRoles role, QComboBox *cb)
{
    auto m = Motors::GetIt();
    int motor = m->getRoleMotor(role);
    
    for (int i=0; i<ARS3000_numConfLines[motor-1]; i++)
    {
        cb->addItem(ARS3000_confLinesText[motor-1][i]);
    }
}

QList<QString> ARS3000::getConfLines(Motors::MotorRoles role)
{
    auto m = Motors::GetIt();
    int motor = m->getRoleMotor(role);
    
    QList<QString> list;
    for (int i=0; i<ARS3000_numConfLines[motor-1]; i++)
    {
        list << ARS3000_confLinesText[motor-1][i];
    }
    return list;
}

void ARS3000::initGUI()
{
    title = tr("ARS 3000 table");
    this->setTitle(title);
    QVBoxLayout *mainVLay = new QVBoxLayout(this);
    
    QHBoxLayout *hLay = new QHBoxLayout();
    mainVLay->addLayout(hLay);
    
    ARS3000_start = new TwoStateButton(":/USBVideo/CameraStart.png",":/USBVideo/CameraStop.png");
    ARS3000_start->setToolTip(tr("Start/stop ars3000 motor"));
    connect(ARS3000_start, SIGNAL(state_changed(bool)), this, SLOT(start(bool)));
    hLay->addWidget(ARS3000_start);
    
    ARS3000_limitterGui = new QCheckBox(tr("Limit"));
    connect(ARS3000_limitterGui, SIGNAL(stateChanged(int)), this, SLOT(confLineChangedSlot()));
    hLay->addWidget(ARS3000_limitterGui);
    
    hLay->addWidget(new QLabel(tr("Number lines")), 0, Qt::AlignRight);
    ARS3000_numConfLinesGui = new Edit(1, 10);
    ARS3000_numConfLinesGui->setImmediateMode();
    ARS3000_numConfLinesGui->setSizePolicy(QSizePolicy::Minimum, QSizePolicy::Fixed);
    connect(ARS3000_numConfLinesGui, SIGNAL(newValue()), this, SLOT(numLinesChanged()));
    connect(ARS3000_numConfLinesGui, SIGNAL(newValue()), this, SLOT(confLineChangedSlot()));
    hLay->addWidget(ARS3000_numConfLinesGui);
    
    for (int b=0; b<2; b++)
    {
        objsl_wid[b] = new QGroupBox(tr("One step data"));
        int margL = objsl_wid[b]->contentsMargins().left();
        objsl_wid[b]->setContentsMargins(margL,0,margL,0);
        auto objMonLay = new QHBoxLayout(objsl_wid[b]);
        margL = objMonLay->contentsMargins().left();
        objMonLay->setContentsMargins(margL,0,margL,0);
        objsl_stepValue[b] = new Edit(0, 1000, 2);
        objsl_speedValue[b] = new Edit(1, 9999, 0);
        objsl_stepValue[b]->setSizePolicy(QSizePolicy::Minimum, QSizePolicy::Fixed);
        objsl_speedValue[b]->setSizePolicy(QSizePolicy::Minimum, QSizePolicy::Fixed);
        connect(objsl_stepValue[b], SIGNAL(newValue()), this, SLOT(confLineChangedSlot()));
        connect(objsl_speedValue[b], SIGNAL(newValue()), this, SLOT(confLineChangedSlot()));
        auto layLeft = new QHBoxLayout();
        layLeft->addWidget(new QLabel(tr("Step: (")+QObject::tr("micro")+tr("m)")));
        layLeft->addWidget(objsl_stepValue[b]);
        auto layRight = new QHBoxLayout();
        layRight->addWidget(new QLabel(tr("Speed: (")+QObject::tr("micro")+tr("m)")), 0, Qt::AlignRight);
        layRight->addWidget(objsl_speedValue[b]);
        objMonLay->addLayout(layLeft);
        objMonLay->addLayout(layRight);
        mainVLay->addWidget(objsl_wid[b]);
        objsl_wid[b]->hide();
    }
    
    QStringList headList;
    headList.append(tr(""));
    headList.append(tr("Name"));
    headList.append(tr("Dist (")+QObject::tr("micro")+tr("m)"));
    ARS3000_TableConf = new QTableWidget(ARS3000_maxNumConfLines, headList.count());
    ARS3000_TableConf->verticalHeader()->setVisible(false);
    ARS3000_TableConf->setDragEnabled(false);
    ARS3000_TableConf->setFocusPolicy(Qt::NoFocus);
    ARS3000_TableConf->horizontalHeader()->setClickable(false);
    ARS3000_TableConf->horizontalHeader()->setMovable(false);
    //ARS3000_TableConf->horizontalHeader()->setResizeMode(QHeaderView::ResizeToContents);
    ARS3000_TableConf->setSortingEnabled(false);
    ARS3000_TableConf->setAlternatingRowColors(true);
    ARS3000_TableConf->setHorizontalHeaderLabels(headList);
    ARS3000_TableConf->horizontalHeader()->setResizeMode(1, QHeaderView::Stretch);
    ARS3000_TableConf->setColumnWidth(0, 30);
    ARS3000_TableConf->setColumnWidth(2, 100);
    mainVLay->addWidget(ARS3000_TableConf);
    
    for (int i=0; i<ARS3000_maxNumConfLines; i++)
    {
        QString stridx = QString::number(i+1);
        ARS3000_radioSelectedConfLine[i] = new QRadioButton(stridx);
        ARS3000_textboxConfLine[i] = new QLineEdit();
        ARS3000_confLinesValuesGui[i] = new Edit(0, 500000);
        
        connect(ARS3000_radioSelectedConfLine[i], SIGNAL(clicked()), this, SLOT(confLineChangedSlot()));
        connect(ARS3000_textboxConfLine[i], SIGNAL(textEdited(QString)), this, SLOT(confLineChangedSlot()));
        connect(ARS3000_confLinesValuesGui[i], SIGNAL(newValue()), this, SLOT(confLineChangedSlot()));
        
        ARS3000_TableConf->setCellWidget(i, 0, ARS3000_radioSelectedConfLine[i]);
        ARS3000_TableConf->setCellWidget(i, 1, ARS3000_textboxConfLine[i]);
        ARS3000_TableConf->setCellWidget(i, 2, ARS3000_confLinesValuesGui[i]);
    }
    this->hide();
}

void ARS3000::numLinesChanged()
{
    for (int r = 0; r < ARS3000_TableConf->rowCount(); r++)
    {
        bool hide = (r >= ARS3000_numConfLinesGui->value());
        ARS3000_TableConf->setRowHidden(r, hide);
    }
}

void ARS3000::loadSettings()
{
    QSettings settings("Param/ARS3000.conf", QSettings::IniFormat);
    
    QList<QVariant> limiters = settings.value("ARS3000_limitters").toList();
    QList<QVariant> numConfLines = settings.value("ARS3000_numConfLines").toList();
    QList<QVariant> selectedConfLines = settings.value("ARS3000_selectedConfLines").toList();
    for (int i=0; i<Motors::numMaxMotors; i++)
    {
        QString stridxI = QString::number(i);
        
        ARS3000_limitter[i] = (i < limiters.count()) ? limiters[i].toInt() : 0;
        ARS3000_numConfLines[i] = (i < numConfLines.count()) ? numConfLines[i].toInt() : 1;
        ARS3000_selectedConfLines[i] = (i < selectedConfLines.count()) ? selectedConfLines[i].toInt() : 0;
        
        QList<QVariant> confLineText = settings.value("ARS3000_confLinesText_" + stridxI).toList();
        QList<QVariant> confLineValue = settings.value("ARS3000_confLinesValue_" + stridxI).toList();
        
        for (int j=0; j<ARS3000_maxNumConfLines; j++)
        {
            QString stridx = QString::number(j+1);
            
            ARS3000_confLinesText[i][j] = (j < confLineText.count()) ? confLineText[j].toString() : (tr("Empty") + stridx);
            ARS3000_confLinesValues[i][j] = (j < confLineValue.count()) ? confLineValue[j].toInt() : 1;
        }
    }
}

void ARS3000::saveSettings()
{
    QSettings settings("Param/ARS3000.conf", QSettings::IniFormat);
    
    QList<QVariant> limiters;
    QList<QVariant> numConfLines;
    QList<QVariant> selectedConfLines;
    
    for (int i=0; i<Motors::numMaxMotors; i++)
    {
        QString stridx = QString::number(i);
        
        limiters.append(ARS3000_limitter[i]);
        numConfLines.append(ARS3000_numConfLines[i]);
        selectedConfLines.append(ARS3000_selectedConfLines[i]);
        
        QList<QVariant> confLineTextList;
        QList<QVariant> confLineValueList;
        for (int j=0; j<ARS3000_maxNumConfLines; j++)
        {
            confLineTextList.append(ARS3000_confLinesText[i][j]);
            confLineValueList.append(ARS3000_confLinesValues[i][j]);
        }
        settings.setValue("ARS3000_confLinesText_" + stridx, confLineTextList);
        settings.setValue("ARS3000_confLinesValue_" + stridx, confLineValueList);
    }
    
    settings.setValue("ARS3000_limitters", limiters);
    settings.setValue("ARS3000_numConfLines", numConfLines);
    settings.setValue("ARS3000_selectedConfLines", selectedConfLines);
    
    settings.setValue("ARS3000_objslmonohr_step", objsl_stepValue[0]->value());
    settings.setValue("ARS3000_objslmonohr_speed", objsl_speedValue[0]->value());
    settings.setValue("ARS3000_objslpmt_step", objsl_stepValue[1]->value());
    settings.setValue("ARS3000_objslpmt_speed", objsl_speedValue[1]->value());
}

void ARS3000::confLineChangedSlot()
{
    int num = currentMotorNum;
    ARS3000_limitter[num] = ARS3000_limitterGui->isChecked();
    ARS3000_numConfLines[num] = ARS3000_numConfLinesGui->value();
    for (int i=0; i<ARS3000_numConfLines[num]; i++)
    {
        ARS3000_confLinesText[num][i] = ARS3000_textboxConfLine[i]->text();
        ARS3000_confLinesValues[num][i] = ARS3000_confLinesValuesGui[i]->value();
        if (ARS3000_radioSelectedConfLine[i]->isChecked())
            ARS3000_selectedConfLines[num] = i;
    }
    saveSettings();
    
    auto m = Motors::GetIt();
    auto role = m->mapVirtToRole(num+1);
    emit confLineChanged(role, ARS3000_selectedConfLines[num]);
}

bool ARS3000::isMotorRoleARS3000(Motors::MotorRoles role)
{
    return (role >= Motors::NDFilter && role <= Motors::ObjYSlPMT);
}

bool ARS3000::init()
{
    return true;
}

bool ARS3000::setSlitWidth(int idx, double width)
{
    if (idx >= 2)
        return false;
    
    auto motors = Motors::GetIt();
    if (!motors)
        return false;

#ifdef _NEW_MOTORS_CHANGES
    int csb = motors->csb_motor_count[0];
    if (csb < 0) csb = 6;
    else csb += motors->csb_motor_count[1] +  motors->csb_motor_count[2];
#else
    int csb = motors->csb_motor_count;
    if (csb < 0) csb = 6;
#endif
    
    int midx0 = 1+idx; //target motor
    int midx = csb+1+midx0;
    midx = motors->mapPhysToVirt(midx);
    
    
    QSettings settings("Param/ars.conf", QSettings::IniFormat);
    double k = settings.value("motor_k/"+QString::number(midx0), 1).toDouble();
    k *= settings.value("slit_k/"+QString::number(idx), 1).toDouble();
    
    double speed = 200;// * k;
    double zeropos = settings.value("slit_zeropos/"+QString::number(idx), 0).toDouble();
    width += zeropos;
    
    if (ars_slit_pos[idx] < 0 || width == ars_slit_pos[idx] || width == 0) {
        ars_slit_pos[idx] = 0;
        motors->goDownE(midx, speed);
        
        SleepExport::msleep_gui(300);
        while (!isReady()) {
            qApp->processEvents();
            SleepExport::msleep2(10);
        }
    }
    
    if (width == ars_slit_pos[idx])
        return true;
    
    motors->go(midx, 1, speed, 1e3 * (width - ars_slit_pos[idx]) * k);
    ars_slit_pos[idx] = width;
    return true;
}

bool ARS3000::setLaser(int idx)
{
    auto motors = Motors::GetIt();
    if (!motors)
        return false;

#ifdef _NEW_MOTORS_CHANGES
    int csb = motors->csb_motor_count[0];
    if (csb < 0) csb = 6;
    else csb += motors->csb_motor_count[1] +  motors->csb_motor_count[2];
#else
    int csb = motors->csb_motor_count;
    if (csb < 0) csb = 6;
#endif
    
    int midx0 = 0; //target motor
    int midx = csb+1+midx0;
    midx = motors->mapPhysToVirt(midx);
    
    
    QSettings settings("Param/ars.conf", QSettings::IniFormat);
    double k = settings.value("motor_k/"+QString::number(midx0), 1).toDouble();
    
    double speed = 700 * k;
    double pos = settings.value("laser_select_pos/"+QString::number(idx), 0).toDouble();
    
    if (ars_laser_pos < 0 || ars_laser_pos == pos || pos == 0) {
        ars_laser_pos = 0;
        motors->goDownE(midx, speed);
        
        SleepExport::msleep_gui(300);
        while (!isReady()) {
            qApp->processEvents();
            SleepExport::msleep2(10);
        }
    }
    
    if (pos == ars_laser_pos)
        return true;
    
    motors->go(midx, 1, speed, 1e3 * (pos - ars_laser_pos) * k);
    ars_laser_pos = pos;
    return true;
}

bool ARS3000::isReady()
{
    auto motors = Motors::GetIt();
    if (!motors)
        return true;

#ifdef _NEW_MOTORS_CHANGES
    int csb = motors->csb_motor_count[0];
    if (csb < 0) csb = 6;
    else csb += motors->csb_motor_count[1] +  motors->csb_motor_count[2];
#else
    int csb = motors->csb_motor_count;
    if (csb < 0) csb = 6;
#endif
    
    for (int i = csb; i < csb + 4, i < 16; i++) { //fixme first condition will be ignored
        if (motors->Motor_on[i])
            return false;
    }
    
    return true;
}

void ARS3000::updateMotors()
{
    emit motorsUpdated();
}
