#include "EG3000.h"
#include <i_eg3000.h>
#include "gui/status.hpp"
#include <QSettings>
#include <QTimer>
#include <QApplication>
#include <QFile>
#include <QDebug>
#include <QAction>
#include "scanplot.h"
#include "modules.hpp"
#include "motors.hpp"
#include "EG3000_Program.h"
#include "eg3000_calibrations.h"

EG3000_CONTROL::EG3000_CONTROL()
{
    need_program = false;

    init_analog_calib_cache();

    prev_usb_stat = -1;
    SWusb1 = new StatusWidget1(tr("USB"));
    SWusb1->label->setToolTip(tr("EG-3000"));
    s_usbStat(EDno);

    QAction *a;
    SWusb1->setContextMenuPolicy(Qt::ActionsContextMenu);
    SWusb1->addAction(a = new QAction(tr("Reconnect"), SWusb1));
    connect(a, SIGNAL(triggered()), this, SLOT(drop_connect()));

    connect(this, SIGNAL(usbStatus(int)), this, SLOT(s_usbStat(int)));
    connect(&rThread, SIGNAL(status(int)), this, SLOT(s_usbStat(int)));

    QTimer *boot_timeout = new QTimer(this);
    boot_timeout->setInterval(2500);
    boot_timeout->setSingleShot(true);
    connect(boot_timeout, SIGNAL(timeout()), this, SLOT(s_boot_timeout()));
    connect(this, SIGNAL(restart_boot_timer()), boot_timeout, SLOT(start()));


    if (sizeof(BYTE) != 1 || sizeof(WORD) != 2 || sizeof(DWORD) != 4 || sizeof(float) != 4)
        qDebug() << "wrong data size" <<
                    sizeof(BYTE) << sizeof(WORD) << sizeof(DWORD) << sizeof(float);
    if (sizeof(EG_CalData) != 228 || sizeof(EG_ForceCurve) != 20 ||
        sizeof(EG_ScanParams) != 108 || sizeof(EG_StepMotorData) != 16)
        qDebug() << "wrong eg_data size" <<
                    sizeof(EG_CalData) << sizeof(EG_ForceCurve) <<
                    sizeof(EG_ScanParams) << sizeof(EG_StepMotorData);
}

void EG3000_CONTROL::program(QString val)
{
    need_program = true;
    path = val;
    emit device_status(2);
}

void EG3000_CONTROL::s_boot_timeout()
{
    boot_timeout = true;
}

bool EG3000_CONTROL::PrepareForReprogramming(void *devHandle, struct eg3000_read_handle &read_handle,
                                             int &prevStatus, bool &need_boot, bool &is_flash_good)
{
    wThread.start_stop(false);
    rThread.start_stop(false);
    prevStatus = EDconn;
    emit usbStatus(prevStatus);
    
    msleep(50);
    
    while (!wThread.write_lock.tryLock(1))
        msleep(10);
    
    //todo make there signal and move that code away
    eg3000_fb_control(devHandle, fbOff); //disable feedback
    eg3000_osc_startstop(devHandle, 0); //stop oscilloscope
    eg3000_status_onoff(devHandle, 0); //disable spm info send
    eg3000_set_head_dac(devHandle, 32767); //set tip voltage ~0
    eg3000_hv_zero(devHandle); //hv outs to 0
    qDebug() << "eg3000 stop";
    
    msleep(500);
    
    eg3000_send_empty_packet(devHandle, ADR_CPU, CPU_GO_REBOOT);
    qDebug() << "eg3000 reset";
    
    wThread.write_lock.unlock();
    
    need_boot = true;
    is_flash_good = true;
    
    msleep(500);
    
    eg3000_clear(devHandle);
    eg3000_clear_read_handle(&read_handle);
    
    return eg3000_is_ok(devHandle);
}

bool EG3000_CONTROL::BootStart(void *devHandle, bool need_boot)
{
    bool st = true;
    rThread.read_lock.lock();
    wThread.write_lock.lock();
    
    if (!need_program)
    {
        //query
        eg3000_bl_stats(devHandle); //ping bootloader
        eg3000_cpu_loopback(devHandle); //ping cpu
        st = eg3000_is_ok(devHandle);
    }
    
    emit restart_boot_timer();
    boot_timeout = false;
    
    if (!(st && need_boot && !m_closeThread)) //todo need_boot is always true here
    { 
        rThread.read_lock.unlock();
        wThread.write_lock.unlock();
    }
    
    return st;
}

void EG3000_CONTROL::ParsePackage_ADR_BOOTLOADER(void *devHandle, struct eg3000_read_handle &read_handle,
                                                 bool &st, bool &is_flash_good, bool &boot_retried, BYTE &boot_timeout_count)
{
    BYTE SubAddress = read_handle.sub_address;
    if (SubAddress == BL_ERROR)
    {
        qDebug() << "bl error" << (void*)((BYTE *)read_handle.buffer)[0];
    }
    if (SubAddress == BL_STATS)
    {
        //unsigned int flash_size = ((DWORD*)read_handle.buffer)[0];
        is_flash_good = ((BYTE *)read_handle.buffer)[4];
        
        if (!need_program)
        {
            if (!is_flash_good)
            {
                if (boot_timeout_count == 0)
                {
                    if (!m_waitMainInit)
                        boot_timeout_count = 4; //skip second warning
                    qDebug() << "no flash in controller!";
                    emit device_status(0);
                }
            }
            else
            { //auto_boot
                st = eg3000_bl_run(devHandle);
                boot_retried = false;
            }
        }
    }
    if (SubAddress == BL_NO_FLASH)
    {
        if (!need_program)
        {
            if (boot_retried)
            {
                if (boot_timeout_count == 0)
                {
                    if (!m_waitMainInit)
                        boot_timeout_count = 4; //skip second warning
                    qDebug() << "no flash in controller!";
                    emit device_status(0);
                }
            }
            else
            {
                boot_retried = true;
                msleep(100);
                st = eg3000_bl_run(devHandle);
            }
        }
    }
    
    if (SubAddress == BL_BOOT_OK)
    { //booted
        st = eg3000_cpu_loopback(devHandle); //ping
    }
}

void EG3000_CONTROL::run()
{
    m_closeThread = false;
    m_waitMainInit = true;
    need_reconnect = false;

    struct eg3000_read_handle read_handle;
    eg3000_init_read_handle(&read_handle);
    rThread.read_handle = &read_handle;
    void *devHandle = nullptr;
    bool st = false;

    bool need_boot = true;
    bool is_flash_good = true;
    boot_timeout = false;
    int prevStatus = -1;
    bool use_4000 = false;
    bool slept = true;

    while (!m_closeThread)
    {
        if (st)
            st = eg3000_is_ok(devHandle);
        if (need_reconnect)
        {
            st = false;
            need_reconnect = false;
        }

        //prepare for reprogramming
        if (st && need_program && !need_boot)
        {
            st = PrepareForReprogramming(devHandle, read_handle, prevStatus, need_boot, is_flash_good);
        }

        //boot start ---------------------------------------
        if (st && need_boot && !m_closeThread)
        {
            st = BootStart(devHandle, need_boot);
        }
        
        bool boot_retried = false;
        BYTE boot_timeout_count = 0;
        while (st && need_boot && !m_closeThread)
        {
            st = eg3000_update_parse(devHandle, &read_handle);
            BYTE Address = read_handle.address;
            
            //packet parsing
            if (read_handle.packet_recieved)
            {
                if (Address == ADR_BOOTLOADER)
                {
                    ParsePackage_ADR_BOOTLOADER(devHandle, read_handle, st, is_flash_good, boot_retried, boot_timeout_count);
                }
                if (Address == ADR_CPU)
                {
                    need_boot = false;
                }

                eg3000_discard_packet(&read_handle);
            }
            else
                msleep(10);

            //program
            if (need_program && need_boot)
            {
                emit device_status(3);
                bool ok = (path.endsWith(".hex") ? program_hex : program_bin)(path, &read_handle, devHandle);
                st = eg3000_is_ok(devHandle);
                if (ok && st)
                {
                    need_program = false;
                    emit device_status(4);
                } else {
                    qDebug() << "retry program";
                    emit device_status(2);
                }

                emit restart_boot_timer();
                boot_timeout = false;
                boot_timeout_count = 0;

                msleep(100);
            }

            //fail recovery
            if (st && (boot_timeout || read_handle.packet_malformed) && need_boot && !m_closeThread && is_flash_good)
            {
                emit restart_boot_timer();
                boot_timeout = false;

                if (read_handle.packet_malformed)
                {
                    eg3000_clear_read_handle(&read_handle);
                    eg3000_send_empty_packet(devHandle, ADR_CPU, CPU_USB_BUF_RST);
                    msleep(100);
                }
                else if (!m_waitMainInit)
                    boot_timeout_count++;

                //resend query
                eg3000_bl_stats(devHandle); //ping bootloader
                eg3000_cpu_loopback(devHandle); //ping cpu
                st = eg3000_is_ok(devHandle);

                if (boot_timeout_count == 3)
                {
                    qDebug() << "boot timeout";
                    emit device_status(1);
                }
            }

            if (!(st && need_boot && !m_closeThread))
            {
                rThread.read_lock.unlock();
                wThread.write_lock.unlock();
            }
        }
        //boot end------------------------------------

        {
            int stat = st ? (need_boot ? EDconn : EDok) : EDno;
            if (!m_waitMainInit && prevStatus != stat)
            {
                wThread.start_stop(st);
                rThread.start_stop(st);
                emit usbStatus(prevStatus = stat);
            }
        }

        //close
        if (!st && devHandle != nullptr)
        {
            //--todo move to function CloseConnection
            rThread.read_lock.lock();
            wThread.write_lock.lock();

            if (eg3000_is_ok(devHandle))
                eg3000_send_empty_packet(devHandle, ADR_CPU, CPU_USB_BUF_RST); //l2k soft reset = need cleanup

            eg3000_close(devHandle);
            devHandle = nullptr;

            rThread.read_lock.unlock();
            wThread.write_lock.unlock();
            //-----------------
        }

        //open
        if (!st && !m_closeThread)
        {
            QSettings settings("Param/controller.conf", QSettings::IniFormat);
            bool use = settings.value("use_eg4000", true).toBool();
            if (use_4000 != use)
                eg3000_set_backend(use ? 1 : 0); //libusb:ftdi
            use_4000 = use;

            need_boot = true;
            is_flash_good = true;

            const WORD pid = PID+(use_4000 ? PID_lastNum : 0);
            struct s_info array;
            WORD size = 0;
            st = eg3000_list(VID, pid, &size, &array, 1);
            if (st && size != 0)
            {
                st = eg3000_open(array.handle, VID, pid, &devHandle);
                rThread.devHandle = devHandle;
                wThread.devHandle = devHandle;
            }

            if (st)
            {
                st = eg3000_clear(devHandle);
                eg3000_clear_read_handle(&read_handle);
                if (slept) {
                    slept = false;
                    continue;
                }
            }
            //---------------
        }
        
        slept = true;
        for (int i = 0; i < (m_waitMainInit ? 1 : 10) && !m_closeThread; i++)
            msleep(50);
    }

    if (st && !need_boot) {
        eg3000_fb_control(devHandle, fbOff); //disable feedback
        eg3000_osc_startstop(devHandle, 0); //stop oscilloscope
        eg3000_status_onoff(devHandle, 0); //disable spm info send
        eg3000_set_head_dac(devHandle, 32767); //set tip voltage ~0
        //eg3000_hv_zero(devHandle); //hv outs to 0
        

        eg3000_send_empty_packet(devHandle, ADR_CPU, CPU_USB_BUF_RST);
    }

    if (devHandle != nullptr)
        eg3000_close(devHandle);
    eg3000_close_read_handle(&read_handle);
}

void EG3000_CONTROL::thread_close()
{
    wThread.start_stop(false);
    rThread.start_stop(false);
    while (wThread.isRunning() && rThread.isRunning()) {
        qApp->processEvents();
        msleep(3);
    }

    m_closeThread = true;
    while (isRunning()) {
        qApp->processEvents();
        msleep(3);
    }
}
void EG3000_CONTROL::drop_connect()
{
    need_reconnect = true;
}

void EG3000_CONTROL::s_usbStat(int idx)
{
    if (prev_usb_stat == idx)
        return;
    prev_usb_stat = idx;

    if (!SWusb1)
        return;
    
    auto status = SWusb1->status;

    switch (prev_usb_stat) {
    case EDno:
        status->set(StatusWidget::Cross, Qt::red);
        status->setToolTip(tr("No device"));
        break;
    case EDconn:
        status->set(StatusWidget::Circle, QColor(255, 128, 0));
        status->setToolTip(tr("Loading"));
        break;
    case EDok:
        status->set(StatusWidget::Circle, Qt::green);
        status->setToolTip(tr("Connected"));
        break;
    case EDplay:
        status->set(StatusWidget::Play, Qt::green);
        status->setToolTip(tr("Working"));
        break;
    }
}

//------------------------------------------------------------------------
USB_WRITE_THREAD::USB_WRITE_THREAD()
{
    want_running = false;
    devHandle = nullptr;
}
void USB_WRITE_THREAD::start_stop(bool st)
{
    if (want_running == st)
        return;
    
    want_running = st;
    if (want_running)
        start(QThread::HighPriority);
}
void USB_WRITE_THREAD::run()
{
    for (int i = 0; i < writers.count(); i++)
        writers.at(i)->resetWriter();

    qDebug() << "eg3000 started";

    while (want_running) {
        if (!write_lock.tryLock(1)) {
            msleep(1);
            continue;
        }

        bool sent = false;
        for (int i = 0; i < writers.count(); i++) {
            if (writers.at(i)->sendData(devHandle))
                sent = true;
            if (writers.at(i)->sendDataIndirect(this))
                sent = true;
        }

        if (!eg3000_is_ok(devHandle))
            want_running = false;

        write_lock.unlock();
        if (!sent)
            msleep(1);
    }
}

BYTE USB_WRITE_THREAD::raw_send(const void *buffer, WORD data_size)
{
    return eg3000_raw_send(devHandle, buffer, data_size);
}
BYTE USB_WRITE_THREAD::send_packet(BYTE Address, BYTE SubAddress, const void *buffer, WORD data_size)
{
    return eg3000_send_packet(devHandle, Address, SubAddress, buffer, data_size);
}

//------------------------------------------------------------------------
USB_READ_THREAD::USB_READ_THREAD()
{
    want_running = false;
    devHandle = nullptr;
    read_handle = nullptr;
}
void USB_READ_THREAD::start_stop(bool st)
{
    if (want_running == st)
        return;
    
    want_running = st;
    if (want_running)
    {
        emit capBits(0, 0); //eg3k
        emit capBits(2, 0); //head
        emit capBits(3, 0); //stage
        emit capBits(4, 0); //z3
        emit capBits(6, 0); //motors

        start(QThread::HighestPriority);
    }
}
void USB_READ_THREAD::run()
{
    DWORD last_bytes_cnt = 0;
    int prevStatus = -1;

    while (want_running) {
        if (!read_lock.tryLock(1)) {
            msleep(1);
            continue;
        }

        if (!eg3000_update_parse(devHandle, read_handle))
            want_running = false;

        bool maybe_have_more_data = (last_bytes_cnt != read_handle->buf_data_length) || read_handle->header_read;
        last_bytes_cnt = read_handle->buf_data_length;

        //packet parsing
        if (read_handle->packet_recieved)
        {
            ParseRecievedPackage(prevStatus);
        }
        else if (!maybe_have_more_data)
            msleep(1);

        read_lock.unlock();
    }
}

void USB_READ_THREAD::ParseRecievedPackage(int &prevStatus)
{
    BYTE Address = read_handle->address;
    BYTE SubAddress = read_handle->sub_address;
    
    for (int i = 0; i < readers.count(); i++)
    {
        bool isPackageReaded = readers.at(i)->processPacket(Address, SubAddress, read_handle->buffer, read_handle->packet_size);
        if (isPackageReaded)
            break;
    }
    
    if (SubAddress == CPU_DEVICE_CAP_BITS &&
        (Address - ADR_HEAD) >= 0 &&
        (Address - ADR_HEAD) <= 8)
    {
        emit capBits(2 + (Address - ADR_HEAD), *((DWORD *)read_handle->buffer));
    }
    else if (Address == ADR_CPU)
    {
        if (SubAddress == CPU_CAP_BITS)
            emit capBits(0, *((DWORD *)read_handle->buffer));
        
        if (SubAddress == CPU_STATUS)
        {
            EG_Status *Deflectomter = (EG_Status *)read_handle->buffer;
            
            int stat = Deflectomter->Scanning != mdIdle ? EDplay : EDok;
            if (prevStatus != stat)
                emit status(prevStatus = stat);
        }
    }
    
    eg3000_discard_packet(read_handle);
}
