#include "eg3000_calibrations.h"

#include <QObject>
#include <QHash>
#include <QSettings>
#include "tools.hpp"

inline
    double linear_calib(double value, double from_min, double from_max, double to_min, double to_max)
{
    return (value - from_min) / (from_max - from_min) * (to_max - to_min) + to_min;
}

static
    QHash<BYTE, double> analog_ch_gain;
double input_gain(BYTE ch)
{
    return analog_ch_gain.value(ch, 1);
}
void input_gain_set(BYTE ch, double gain)
{
    analog_ch_gain.insert(ch, gain);
}
void input_gain_clear(BYTE ch)
{
    analog_ch_gain.remove(ch);
}

static
    QHash<BYTE, double> analog_ch_shift;
double input_shift(BYTE ch)
{
    return analog_ch_shift.value(ch, 0);
}
void input_shift_set(BYTE ch, double shift)
{
    analog_ch_shift.insert(ch, shift);
}
void input_shift_clear(BYTE ch)
{
    analog_ch_shift.remove(ch);
}

//from [a..b] to [a0..b0]
double analog_calib(double value, const char *name, double a0, double b0)
{
    QSettings settings("Param/analog.conf", QSettings::IniFormat);
    double a = settings_double(settings, QString() + name + "_min", a0);
    double b = settings_double(settings, QString() + name + "_max", b0);
    return linear_calib(value, a, b, a0, b0);
}

double analog_calib_inverse(double value, const char *name, double a0 = -10., double b0 = 10., double a = -10., double b = 10.)
{
    QSettings settings("Param/analog.conf", QSettings::IniFormat);
    a = settings_double(settings, QString() + name + "_min", a);
    b = settings_double(settings, QString() + name + "_max", b);
    return linear_calib(value, a0, b0, a, b);
}

//

static
    QMap<int, QPair<double, double> > analog_map_cache;
void init_analog_calib_cache()
{
    analog_map_cache.clear();
    
    QSettings settings("Param/analog.conf", QSettings::IniFormat);
    
    QList<const char *> lst;
    QList<int> key;
    lst << "adc/v1" << "adc/v1mag" << "adc/v2" << "adc/v2mag" << "adc/v3" << "adc/v4" << "adc/v5" << "adc/v1hr" << "adc/v2hr" << "dac/vtip";
    key << 1 << 11 << 2 << 12 << 3 << 4 << 5 << 21 << 22 << 101;
    for (int i = 0; i < lst.size() && i < key.size(); i++) {
        double a = settings_double(settings, QString() + lst.at(i) + "_min", -10);
        double b = settings_double(settings, QString() + lst.at(i) + "_max", 10);
        
        analog_map_cache.insert(key.at(i), qMakePair(a, b));
    }
}

QPair<double, double> GetPairAnalogMapCache(int key)
{
    return analog_map_cache.value(key, qMakePair(-10., 10.));
}

double input_calib(BYTE ch, double val)
{
    double gain = input_gain(ch);
    double shift = input_shift(ch);
    
    //bool unwrap_unsigned = false;
    int idx = 0;
    switch (ch) {
    default:
        return (linear_calib(val, 0, 65535, -10, 10) + shift) * gain;
        break;
    case srcDACX:
    case srcDACY:
    case srcDACZ:
    case srcDACX2:
    case srcDACY2:
    case srcDACZ2:
    case srcDACZ3:
        return (shift + gain * val);
        break;
    case srcADC1:
    case srcDFL:
        idx = 1;
        break;
    case srcMag1:
        //unwrap_unsigned = true;
    case srcMagCos1:
    case srcMagSin1:
        idx = 11;
        break;
    case srcADC2:
    case srcLFL:
        idx = 2;
        break;
    case srcMag2:
        //unwrap_unsigned = true;
        idx = 12;
        break;
    case srcADC3:
    case srcIntensity:
        idx = 3;
        break;
    case srcADC4:
        idx = 4;
        break;
    case srcADC5:
        idx = 5;
        break;
    case srcADCHR1:
        idx = 21;
        break;
    case srcADCHR2:
        idx = 22;
        break;
    case srcVTip:
        idx = 101;
        break;
    }
    
    //fix negative part to be at top half for sign-scrambled unsigned values
    //if (unwrap_unsigned && val < 32768)
    //    val += 65536;
    
    QPair<double, double> p = analog_map_cache.value(idx, qMakePair(-10., 10.));
    return (linear_calib(val, 0., 65535., p.first, p.second) + shift) * gain;
}


QHash<BYTE, QString> analog_ch_unit;
QString input_units(BYTE ch)
{
    return analog_ch_unit.value(ch, QObject::tr("V"));
}
void input_unit_set(BYTE ch, QString unit)
{
    analog_ch_unit.insert(ch, unit);
}
void input_unit_clear(BYTE ch)
{
    analog_ch_unit.remove(ch);
}
