#include "modules.hpp"
#include <QDebug>
#include <ttabwidget.hpp>
#include "options_window.h"
#include <QHBoxLayout>
#include "widgets/threedoverlaybuttons.h"
#include "widgets/twodoverlaybuttons.h"
#include "IExternalScan.h"
#include "2d/ovcontrol_2d.h"
#include "ars3000.h"
#include "singleapp.h"

#include "modules/scanmodule.h"
#include "modules/optionsmodule.h"
#include "modules/eg3000module.h"
#include "modules/datamodule.h"
#include "modules/ccdmatrixmodule.h"
#include "modules/ccdplotmodule.h"
#include "modules/spmmodule.h"
#include "modules/videomodule.h"
#include "modules/oscilloscopemodule.h"
#include "modules/motorsmodule.h"
#include "modules/samplecontrolmodule.h"

ModuleManager *ModuleManager::instance = nullptr;

I_MODULE::~I_MODULE()
{
    if (ModuleManager::instance)
        ModuleManager::instance->unload(this);
}

ModuleManager::ModuleManager()
{
    if (instance)
        throw instance;
    instance = this;
    isRunning = false;

    ModuleManager::instance->addToFactoryGui<OptionsModule>();
    ModuleManager::instance->addToFactory<EG3000Module>();
    ModuleManager::instance->addToFactoryGui<DataModule>();
    ModuleManager::instance->addToFactory<CCDMatrixModule>();
    ModuleManager::instance->addToFactoryGui<CCDPlotModule>();
    ModuleManager::instance->addToFactoryGui<ScanModule>();
    ModuleManager::instance->addToFactoryGui<SPMModule>();
    ModuleManager::instance->addToFactoryGui<MotorsModule>();
    ModuleManager::instance->addToFactoryGui<VideoModule>();
    ModuleManager::instance->addToFactoryGui<OscilloscopeModule>();
    ModuleManager::instance->addToFactoryGui<SampleControlModule>();
}

ModuleManager::~ModuleManager()
{
    instance = nullptr;

    if (isRunning)
        foreach (I_MODULE *m, modules)
            m->setRunning(false);

    foreach (I_MODULE *m, modules)
        delete m;

    modules.clear();
}

I_MODULE *ModuleManager::load(const QLatin1String name, QWidget *parent)
{
    if (!factory.contains(name))
        return nullptr;

    QList<QLatin1String> l;
    if (factory[name].requires)
        l = factory[name].requires();
    for (int i = 0; i < l.size(); i++)
        if (!modules.contains(l.at(i)))
            if (!load(l.at(i), parent))
                qCritical() << "can't create module" << QString(l.at(i)) << "required for" << QString(name);

    I_MODULE *m;
    if (factory[name].creategui)
        m = factory[name].creategui(parent);
    else
        m = factory[name].create();
    add(name, m);
    return m;
}

void ModuleManager::add(const QLatin1String name, I_MODULE *module)
{
    QMapIterator<QLatin1String, I_MODULE *> i(modules);
    while (i.hasNext()) {
        i.next();
        i.value()->connected(name, module);
        module->connected(i.key(), i.value());
    }

    if (isRunning)
        module->setRunning(true); //TODO fix order for dependencies

    modules.insert(name, module);
}

void *ModuleManager::find1VoidPtr(const QLatin1String name) const noexcept
{
    I_MODULE *module = ModuleManager::instance->find1(name);

    if (module == nullptr)
        return nullptr;

    return module->basecast();
}

void *ModuleManager::find1Safety(const QLatin1String name, bool throwExeption) const
{
    auto module = find1VoidPtr(name);

    if (module == nullptr)
    {
        if (!throwExeption)
            return nullptr;

        //assert("Error 0x0001. \"" + name + "\" is not exists or not inited yet!");
        //throw "Error 0x0001. \"" + name + "\" is not exists or not inited yet!";
        qFatal(QString("Error 0x0001. \"" + name + "\" is not exists or not inited yet!").toLatin1());
    }

    return module;
}

I_MODULE *ModuleManager::find1(const QLatin1String name) const
{
    return modules.value(name);
}
const QList<I_MODULE *> ModuleManager::find(const QLatin1String name) const
{
    return modules.values(name);
}

const QLatin1String ModuleManager::find(I_MODULE *module) const
{
    foreach (QLatin1String name, modules.keys(module))
        return name;

    return QLatin1String(nullptr);
}

void ModuleManager::unload(const QLatin1String name)
{
    QList<I_MODULE *> l = modules.values(name);
    if (isRunning)
        foreach (I_MODULE *m0, l)
            m0->setRunning(false);

    foreach (I_MODULE *m0, l)
        foreach (I_MODULE *m, modules)
            if (m != m0)
                m->disconnected(name, m0);

    modules.remove(name);

    foreach (I_MODULE *m0, l)
        delete m0;
}

void ModuleManager::unload(I_MODULE *module)
{
    QLatin1String name = find(module);
    if (!name.latin1())
        return;

    if (isRunning)
        module->setRunning(false);

    foreach (I_MODULE *m, modules)
        if (m != module)
            m->disconnected(name, module);

    modules.remove(name, module);
}

void ModuleManager::setRunning(bool running)
{
    if (isRunning == running)
        return;

    isRunning = running;

    foreach (I_MODULE *m, modules)
        m->setRunning(isRunning);
}
