#include "relayoptions.h"

#include <QLayout>
#include <QGroupBox>
#include <QHeaderView>
#include <QCheckBox>
#include <QLabel>
#include <QComboBox>
#include "tools.hpp"
#include "QFile"
#include <QSettings>
#include "tools.hpp"
#include "modules.hpp"
#include <QtDebug>
#include "modules/relaymodule.h"

extern "C" {
///
/// \brief relay_set
/// \param on 1 = on, 0 = off, -1 = skip
/// \param releId >1 = single rele, -1 = all
/// \return 
///
int relay_set(int on, int releId);
}

RelayOptions::RelayOptions(QWidget *parent)
{
    numberRelay = relayCount();
    
    LoadRoles();
    disableAll();
    InitGUI();
}

void RelayOptions::LoadRoles()
{
    //See "enum StandartRelayRoles" for roles list
    strRoles.append(tr("None"));
    strRoles.append(tr("Controller"));
    
    QFile file("ReleRoles.txt");
    if (!file.open(QIODevice::ReadOnly))
        return;
    
    while (!file.atEnd())
    {
        auto line = file.readLine();
        
        line.replace('\n', "");
        line.replace('\r', "");
        if (line.isEmpty())
            continue;

        strRoles.append(line);
    }
}

void RelayOptions::InitGUI()
{
    mainWidget = new QWidget();
    auto mainLay = new QVBoxLayout(mainWidget);
    
    InitReleTable(mainLay);
}

void RelayOptions::InitReleTable(QVBoxLayout *mainLay)
{
    QGroupBox *releGB = new QGroupBox(tr("Rele table"));
    QVBoxLayout *releLay = new QVBoxLayout(releGB);
    
    mainLay->addWidget(releGB);
    
    QStringList headList;
    headList.append(tr("#"));
    headList.append(tr("On"));
    headList.append(tr("Role"));
    
    TableRele = new QTableWidget(maxNumRele, headList.count());
    releLay->addWidget(TableRele);
    
    TableRele->verticalHeader()->setVisible(false);
    TableRele->setDragEnabled(false);
    TableRele->setFocusPolicy(Qt::NoFocus);
    TableRele->horizontalHeader()->setClickable(false);
    TableRele->horizontalHeader()->setMovable(false);
    TableRele->horizontalHeader()->setResizeMode(0, QHeaderView::ResizeToContents);
    TableRele->horizontalHeader()->setResizeMode(1, QHeaderView::ResizeToContents);
    TableRele->setSortingEnabled(false);
    TableRele->setAlternatingRowColors(true);
    TableRele->setHorizontalHeaderLabels(headList);
    TableRele->setColumnWidth(2, 200);
    
    QSettings settings("Param/rele.conf", QSettings::IniFormat);
    
    for (int i = 0; i < maxNumRele; i++)
    {
        QString stridx = QString::number(i + 1);
        
        auto numLabel = new QLabel(stridx);
        auto onOffSwitcher = new QCheckBox();
        cbPowerOn.append(onOffSwitcher);
        
        auto role = new QComboBox();
        role->setStyleSheet(DEFAULT_COMBO_BOX_STYLE);
        role->addItems(strRoles);
        int currRoleIndex = settings.value("role/" + stridx, 0).toInt();
        if (currRoleIndex < strRoles.count())
            role->setCurrentIndex(currRoleIndex);
        cbRoles.append(role);
        
        TableRele->setCellWidget(i, 0, centerWidget(numLabel));
        TableRele->setCellWidget(i, 1, centerWidget(onOffSwitcher));
        TableRele->setCellWidget(i, 2, role);
        
        connect(onOffSwitcher, SIGNAL(toggled(bool)), this, SLOT(relayPowerSwitched(bool)));
        connect(role, SIGNAL(currentIndexChanged(int)), this, SLOT(saveChanges()));
    }
    
    updateReles();
}

void RelayOptions::updateReles()
{
    currNumRele = relayCount();
    
    for (int i = 0; i < maxNumRele; i++)
    {
        bool showLine = (i < currNumRele);
        
        TableRele->setRowHidden(i, !showLine);
    }
}

void RelayOptions::saveChanges()
{
    QSettings settings("Param/rele.conf", QSettings::IniFormat);
    
    for (int i=0; i<numberRelay; i++)
    {
        QString stridx = QString::number(i + 1);
        
        settings.setValue("role/" + stridx, cbRoles.at(i)->currentIndex());
    }
}

void RelayOptions::relayPowerSwitched(bool on)
{
    auto switcher = static_cast<QCheckBox*>(sender());
    int id = cbPowerOn.indexOf(switcher);
    if (id == -1)
        return;
    
    relay_set(on, id + 1);
}

void RelayOptions::enableController()
{
    enable(RR_Controller);
}

void RelayOptions::disableController()
{
    disable(RR_Controller);
}

void RelayOptions::enable(int role)
{
    SetRelePower(true, role);
}

void RelayOptions::disable(int role)
{
    SetRelePower(false, role);
}

void RelayOptions::SetRelePower(bool on, int role)
{
    if (role == RR_None)
        return;
    
    for (auto cb : cbRoles)
    {
        if (cbRoles.indexOf(cb) >= currNumRele)
            break;
        
        if (cb->currentIndex() == role)
            relay_set(on, cbRoles.indexOf(cb) + 1);
    }
}

int RelayOptions::relayCount()
{
    return relay_set(-1, -1);
}

void RelayOptions::enableAll()
{
    relay_set(1, -1);
}

void RelayOptions::disableAll()
{
    relay_set(0, -1);
}

RelayOptions* RelayOptions::GetIt() noexcept
{
    auto relay = ModuleManager::instance->find1VoidPtr(RelayModule::name());
    return static_cast<RelayOptions*>(relay);
}
