#include "oscilloscope.h"

#include <QSettings>
#include <QMessageBox>
#include <QCloseEvent>
#include <QHBoxLayout>
#include <QStackedLayout>
#include <QCheckBox>
#include <QTimer>
#include "remote_nst_dataid.hpp"
#include "gui/status.hpp"
#include "autocalib.hpp"
#include "xyzcalibrations.hpp"
#include <QDebug>
#include <QtGlobal>
#include "modules.hpp"
#include "scanplot.h"
#include <qwt_plot_canvas.h>
#include "mobile/remote_mobile.hpp"
#include <QVariant>
#include "hardware/eg3000_calibrations.h"
#include "hardware/EG3000.h"
#include "spmtunning.h"
#include "options_window.h"
#include "qwt_plot_layout.h"
#include "2d/plot_2dplot.h"
#include "2d/plot_2dpure.h"
#include "datalistoverlaybuttons.h"
#include "widgets/twodoverlaybuttons.h"
#include "2d/ovcontrol_2d.h"
#include "dacscontrol.h"

void Oscilloscope::GUICreate()
{
    btnShowHideRightMenuBar = new QPushButton(">>");
    btnShowHideRightMenuBar->setToolTip(tr("Show/hide calibration panel"));
    connect(btnShowHideRightMenuBar, SIGNAL(clicked(bool)), this, SLOT(sett_showhide()));
    btnShowHideRightMenuBar->hide();
    
    QHBoxLayout *MainLayout = new QHBoxLayout();
    
    OscStackedWidget = new QStackedWidget(this);
    OscStackedWidget->setContentsMargins(0, 0, 0, 0);
    
    PBCalibMode = new TwoStateButton(":/Osc_window/CalibOf.png",":/Osc_window/CalibOn.png");
    PBCalibMode->setToolTip(tr("Show/hide calibration window"));
    PBCalibMode->hide();
    connect(PBCalibMode, SIGNAL(state_changed(bool)), SLOT(calib_plot_change(bool)));
    
    BToCalib = new QPushButton(makeHugeIcon(":/Osc_window/CalibOf.png"), QString());
    BToCalib->setToolTip(tr("Switch to calibrations"));
    connect(BToCalib, SIGNAL(clicked()), PBCalibMode, SIGNAL(clicked()));
    
    DacsWidget = new DacsControl();
    
    CalibWidget = new XYZCalibrations(this, this);
    connect(CalibWidget, SIGNAL(input_calib_updated()), this, SLOT(updateRelativeState()));
    
    MainLayout->addWidget(OscStackedWidget, 1);
    MainLayout->addWidget(CalibWidget);
    mainWidget = new QWidget;
    mainWidget->setLayout(MainLayout);
    
    
    QWidget *leftwid = new QWidget(OscStackedWidget);
    QHBoxLayout *leftl = new QHBoxLayout(leftwid);
    leftl->setContentsMargins(0, 0, 0, 0);
    
    acp = new AutoCalibPlot(OscStackedWidget);
    acp->osc = this;
    
    OscStackedWidget->addWidget(leftwid);
    OscStackedWidget->addWidget(acp);
    
    QVBoxLayout *VBLLeft = new QVBoxLayout;
    VBLLeft->setAlignment(Qt::AlignTop);
    VBLLeft->setContentsMargins(0, 0, 0, 0);
    
    QVBoxLayout *VBLMid = new QVBoxLayout;
    VBLMid->setSpacing(3);
    VBLMid->setContentsMargins(0, 0, 6, 0);
    
    leftl->addLayout(VBLLeft);
    leftl->addLayout(VBLMid, 1);
    
    DacsWidget->setParent(leftwid);
    connect(CalibWidget, SIGNAL(sig_XYZ_range(double,double,double)), DacsWidget, SLOT(sl_xyz(double,double,double)));
    connect(CalibWidget, SIGNAL(sig_XYZ_range2(double,double,double)), DacsWidget, SLOT(sl_xyz2(double,double,double)));
    connect(CalibWidget, SIGNAL(sig_XYZ_range3(double,double,double)), DacsWidget, SLOT(sl_xyz3(double,double,double)));
    
    MyPl = new plot::T2dPlot();
    MyPl->Init();
    MyPl->SetGridVisible(true);
    MyPl->SetAutoScalingEnabled(false);
    MyPl->Pure2dPlot->XAxisUnit.SetDescAndSIUnit(tr("time"), tr("s"));
    MyPl->Pure2dPlot->YAxisUnit.SetDescription(tr("None"));
    MyPl->Pure2dPlot->SetAllXScales(0, 10);
    MyPl->Pure2dPlot->SetAllYScales(-11, 11);
    MyPl->SetTitle(tr("Osc curve"));
    MyPl->layout()->setContentsMargins(0, 0, 0, 0);
    MyPl->CreateOverlayButtons();
    MyPl->overlayButtons->control->removeFilterButton();
    MyPl->overlayButtons->control->B2DGrid->change_state_man(true);
    MyPl->overlayButtons->setButtonActive(MyPl->overlayButtons->control->B2DGrid, true);
    MyPl->overlayButtons->control->setVisibleDataFitSizeButton(true);
    QSettings settings2("Param/gui.conf", QSettings::IniFormat);
    if (settings2.value("gui/Ratis_xyz_stage_mode", false).toBool())
        MyPl->overlayButtons->control->setShowingSendToDataButton(false);
    MyPl->PlotControl->setVisible(false);
    
    VBLMid->addWidget(MyPl, 1);
    VBLMid->addWidget(DacsWidget);
    
    
    QHBoxLayout *BLayout = new QHBoxLayout;
    BLayout->setSpacing(6);
    BLayout->setAlignment(Qt::AlignLeft);
    
    VBLLeft->addLayout(BLayout);
    
    
    //----------------
    
    start = new TwoStateButton(":start.png",":stop.png");
    disconnect(start, SIGNAL(clicked()), nullptr, nullptr); //wtf that doing?
    start->setToolTip(tr("Start/stop oscilloscope"));
    connect(start, SIGNAL(pressed()), start, SLOT(change_state()));
    connect(start, SIGNAL(state_changed(bool)), this, SLOT(osc_Start(bool)));
    
    if (settings2.value("gui/Ratis_xyz_stage_mode", false).toBool())
        osc_Start(true);
    
    MyPl->overlayButtons->setButtonStyle(start);
    MyPl->overlayButtons->addLeftButton(start);
    
    ePeriod = new Edit(1, 60);
    ePeriod->setImmediateMode();
    ePeriod->setValue(10);
    ePeriod->setStyleSheet("QDoubleSpinBox {border-width:0; background-color:rgba(255,255,255,25);}");
    connect(ePeriod, SIGNAL(newValue(int)), this, SLOT(s_setT(int)));
    SBTLabel = new QLabel(tr("T, sec"));
    MyPl->overlayButtons->allWidgets.append(SBTLabel);
    MyPl->overlayButtons->allWidgets.append(ePeriod);
    
    BShowRelative = new QPushButton(makeHugeIcon(":/plot/relativeScale.png"), QString());
    BShowRelative->setCheckable(true);
    BShowRelative->setToolTip(tr("SGM scale"));
    MyPl->overlayButtons->setButtonStyle(BShowRelative);
    MyPl->overlayButtons->addLeftButton(BShowRelative);
    connect(BShowRelative, SIGNAL(toggled(bool)), SLOT(updateRelativeState()));
    bshowrelative = false;
    
    GrBChannel[0] = new QGroupBox();
    GrBChannel[1] = new QGroupBox();
    GrBChannel[2] = new QGroupBox();
    GrBChannel[3] = new QGroupBox();
    LEdColour[0] = new TRectWidget(Qt::red);
    LEdColour[1] = new TRectWidget(QColor(0, 170, 0));
    LEdColour[2] = new TRectWidget(Qt::blue);
    LEdColour[3] = new TRectWidget(Qt::black);
    const QString words[numChanels] = {"A", "B", "C", "D"};
    for (int i = 0; i < numChanels; i++)
    {
        LChnLabelName[i] = new QLabel(tr("Channel ") + words[i]);
        setGBChannelSelected(i, false);
        CBChannel[i] = new QComboBox(GrBChannel[i]);
        fillCBchannels_(CBChannel[i]);
        
        LAverage[i] = new QLabel(tr("Average"));
        LStDev[i] = new QLabel(tr("RMS"));
        LAverage[i]->hide();
        LStDev[i]->hide();
        
        LEdAverage[i] = new QLineEdit;
        LEdAverage[i]->hide();
        LEdAverage[i]->setReadOnly(true);
        LEdAverage[i]->setFocusPolicy(Qt::NoFocus);
        
        LEdStDev[i] = new QLineEdit;
        LEdStDev[i]->hide();
        LEdStDev[i]->setReadOnly(true);
        LEdStDev[i]->setFocusPolicy(Qt::NoFocus);
        
        GrBChannel[i]->setFocusPolicy(Qt::StrongFocus);
        GrBChannel[i]->setFocusProxy(CBChannel[i]);
        GrBChannel[i]->installEventFilter(this);
        CBChannel[i]->installEventFilter(this);
        
        
        connect(LEdColour[i], SIGNAL(clicked()), this, SLOT(colorClick()));
        
        VBLLeft->addSpacerItem(new QSpacerItem(12, 12, QSizePolicy::Maximum, QSizePolicy::Maximum));
        VBLLeft->addWidget(GrBChannel[i], 1);
        
        QVBoxLayout *VBLMainForGrB = new QVBoxLayout(GrBChannel[i]);
        VBLMainForGrB->setContentsMargins(6, 3, 6, 3);
        
        QVBoxLayout *VBLAverage = new QVBoxLayout;
        VBLAverage->setContentsMargins(0, 0, 0, 0);
        
        QVBoxLayout *VBLStDev = new QVBoxLayout;
        VBLStDev->setContentsMargins(0, 0, 0, 0);
        
        VBLAverage->addWidget(LAverage[i]);
        VBLAverage->addWidget(LEdAverage[i]);
        
        VBLStDev->addWidget(LStDev[i]);
        VBLStDev->addWidget(LEdStDev[i]);
        
        VBLMainForGrB->addLayout(VBLAverage);
        VBLMainForGrB->addLayout(VBLStDev);
        VBLMainForGrB->addWidget(CBChannel[i]);
        LChnLabelName[i]->setAlignment(Qt::AlignCenter);
        LChnLabelName[i]->setStyleSheet("font-weight: bold; color: " + LEdColour[i]->m_Colour.name() + ";");
        VBLMainForGrB->addWidget(LChnLabelName[i], 0, Qt::AlignBottom);
        VBLMainForGrB->addWidget(LEdColour[i], 0, Qt::AlignBottom);
    }
    
    setGBChannelSelected(0, true);
    
    connect(MyPl->overlayButtons, SIGNAL(replacedGUIPlotResized(QSize)), this, SLOT(guiPlotResized(QSize)));
    connect(MyPl->overlayButtons, SIGNAL(replacedGUIPlotSetVisible(bool)), this, SLOT(guiPlotSetVisible(bool)));
    MyPl->overlayButtons->plot2DResized();
}

void Oscilloscope::guiPlotResized(QSize newSize)
{
    auto overlay = MyPl->overlayButtons;
    int downY = overlay->buttonRect.bottom();
    int cbSize = 60;
    auto font = QFont(SBTLabel->font().family(), cbSize*0.45);
    SBTLabel->setScaledContents(true);
    SBTLabel->setFont(font);
    SBTLabel->adjustSize();
    
    ePeriod->setFixedSize(100, cbSize);
    ePeriod->setFont(font);
    int x = newSize.width() - ePeriod->width() - 10;
    int y = downY - 10 - ePeriod->height();
    ePeriod->move(x, y);
    x -= SBTLabel->width() + 5;
    SBTLabel->move(x, downY - ePeriod->height()*0.5 - SBTLabel->height()*0.75);
}

void Oscilloscope::guiPlotSetVisible(bool visible)
{
    ePeriod->setParent(MyPl);
    SBTLabel->setParent(MyPl);
    bool visibleBtns = visible && (MyPl->overlayButtons->visMode == plot::TwoDOverlayButtons::svm_none);
    MyPl->overlayButtons->setButtonVisible(ePeriod, visibleBtns);
    MyPl->overlayButtons->setButtonVisible(SBTLabel, visibleBtns);
}

void Oscilloscope::GUIRefresh()
{
    if (refreshRateLimit.isValid() && refreshRateLimit.elapsed() < 40 + 4 * oscMaxLength / oscRate)
        return;
    
    auto scan = ScanPlot::GetIt();
    
    refreshRateLimit.start();
    
    if (MyPl->Pure2dPlot->singleAutoScale)
        MyPl->Pure2dPlot->QPlot->canvas()->update(); //OverlayButtons flickering bug fix
    else
        MyPl->Pure2dPlot->RefreshInc();
    
    for (int i = 0; i < numChanels && i < curveData.size(); i++) {
        if (!osc_data_idx[i]) continue;
        
        plot::TSIUnit tunit;
        tunit.SetSIUnit(input_units(bshowrelative ? srcNone : osc_data_idx[i]));
        
        double vavg = avg(curveData.at(i)->YSeq) / k_m_nm[i];
        tunit.UpdateLegend(vavg);
        LEdAverage[i]->setText(QString::number(vavg / tunit.Multiplier, 'f', 3));
        LAverage[i]->setText(tr("Average, ") + tunit.Legend);
        
        double vrms = rms(curveData.at(i)->YSeq) / k_m_nm[i];
        tunit.UpdateLegend(vrms);
        LEdStDev[i]->setText(QString::number(vrms / tunit.Multiplier, 'f', 3));
        LStDev[i]->setText(tr("RMS, ") + tunit.Legend);
        
        if (osc_data_idx[i] == srcHeight && tunit.Multiplier != 0.)
        {
            emit newOscHeightRMS(vrms / tunit.Multiplier, tunit.Legend);
            if (auto mob = MobileServer::GetIt()) {
                if (mob->any_need_data("deflect_need"))
                {
                    auto list = QVariantList();
                    list << vrms / tunit.Multiplier;
                    list << tunit.Legend;
                    mob->sendData(DataID::Deflect_zNoise, list);
                }
            }
        }
    }
    
    //todo REFACTOR. move that code away.
    if (scan)
    {
        scan->deviceSchemeWid->MagText->setValue(LEdAverage[0]->text().toDouble());
        scan->deviceSchemeWid->MagSinText->setValue(LEdAverage[1]->text().toDouble());
        scan->deviceSchemeWid->MagCosText->setValue(LEdAverage[2]->text().toDouble());
        scan->deviceSchemeWid->PhaseText->setValue(LEdAverage[3]->text().toDouble());
    }
    
    emit oscRefreshed();
}
