#include "pumpsstage.h"
#include <QSettings>
#include "motors.hpp"
#include <QtDebug>
#include <QComboBox>
#include <QCheckBox>
#include <QRadioButton>

PumpsStage::PumpsStage()
{
    loadSettings();
    initGUI();
}

void PumpsStage::setMotors(Motors *m)
{
    connect(m, SIGNAL(motorStatusChanged(int, bool, bool)), this, SLOT(motorEvent(int, bool, bool)));
    //connect(m, SIGNAL(motorSendingStatusEnded()), this, SLOT(updateQueue()));
}

void PumpsStage::loadSettings()
{
    QSettings settings("Param/motor.conf", QSettings::IniFormat);
    
    settings.beginGroup("PumpStages");
    
    xyStagePosition[0][0] = settings.value("xStage_pos1").toInt();
    xyStagePosition[0][1] = settings.value("xStage_pos2").toInt();
    xyStagePosition[1][0] = settings.value("yStage_pos1").toInt();
    xyStagePosition[1][1] = settings.value("yStage_pos2").toInt();
    
    xyStageLimit[0] = settings.value("xStage_limit").toInt();
    xyStageLimit[1] = settings.value("yStage_limit").toInt();
}

void PumpsStage::saveSettings()
{
    QSettings settings("Param/motor.conf", QSettings::IniFormat);
    
    settings.beginGroup("PumpStages");
    
    settings.setValue("xStage_pos1", xyStagePosition[0][0]);
    settings.setValue("xStage_pos2", xyStagePosition[0][1]);
    settings.setValue("yStage_pos1", xyStagePosition[1][0]);
    settings.setValue("yStage_pos2", xyStagePosition[1][1]);
    
    settings.setValue("xStage_limit", (int)xyStageLimit[0]);
    settings.setValue("yStage_limit", (int)xyStageLimit[1]);
}

void PumpsStage::initGUI()
{
    title = tr("Pumps control ");
    this->setTitle(title);
    QVBoxLayout *mainVLay = new QVBoxLayout(this);
    
    QHBoxLayout *hLay = new QHBoxLayout();
    mainVLay->addLayout(hLay);
    
    startMotorBtn = new TwoStateButton(":/USBVideo/CameraStart.png",":/USBVideo/CameraStop.png");
    startMotorBtn->setToolTip(tr("Start/stop stage motor"));
    startMotorBtn->setMaximumWidth(100);
    connect(startMotorBtn, SIGNAL(state_changed(bool)), this, SLOT(start(bool)));
    hLay->addWidget(startMotorBtn);
    
    limitterCb = new QCheckBox(tr("Limit"));
    connect(limitterCb, SIGNAL(stateChanged(int)), this, SLOT(confLineChangedSlot()));
    hLay->addWidget(limitterCb);
    
    QStringList headList;
    headList.append(tr(""));
    headList.append(tr("Name"));
    headList.append(tr("Dist (")+QObject::tr("micro")+tr("m)"));
    tableConf = new QTableWidget(2, headList.count());
    tableConf->verticalHeader()->setVisible(false);
    tableConf->setDragEnabled(false);
    tableConf->setFocusPolicy(Qt::NoFocus);
    tableConf->horizontalHeader()->setClickable(false);
    tableConf->horizontalHeader()->setMovable(false);
    //tableConf->horizontalHeader()->setResizeMode(QHeaderView::ResizeToContents);
    tableConf->setSortingEnabled(false);
    tableConf->setAlternatingRowColors(true);
    tableConf->setHorizontalHeaderLabels(headList);
    tableConf->horizontalHeader()->setResizeMode(1, QHeaderView::Stretch);
    tableConf->setColumnWidth(0, 30);
    tableConf->setColumnWidth(2, 100);
    mainVLay->addWidget(tableConf);
    
    const QString lineTexts[2] = {tr("Distance 1"), tr("Distance 2")};
    
    for (int i=0; i<2; i++)
    {
        QString stridx = QString::number(i+1);
        radioSelectedConfLine[i] = new QRadioButton(stridx);
        textboxConfLine[i] = new QLineEdit();
        confLinesValuesGui[i] = new Edit(0, 500000);
        
        textboxConfLine[i]->setText(lineTexts[i]);
        textboxConfLine[i]->setReadOnly(true);
        
        connect(radioSelectedConfLine[i], SIGNAL(clicked()), this, SLOT(confLineChangedSlot()));
        connect(confLinesValuesGui[i], SIGNAL(newValue()), this, SLOT(confLineChangedSlot()));
        
        tableConf->setCellWidget(i, 0, radioSelectedConfLine[i]);
        tableConf->setCellWidget(i, 1, textboxConfLine[i]);
        tableConf->setCellWidget(i, 2, confLinesValuesGui[i]);
    }
    this->hide();
}

void PumpsStage::confLineChangedSlot()
{
    int num = currentMotorNum;
    xyStageLimit[num] = limitterCb->isChecked();
    xyStagePosition[num][0] = confLinesValuesGui[0]->value();
    xyStagePosition[num][1] = confLinesValuesGui[1]->value();
    
    saveSettings();
}

void PumpsStage::motorEvent(int motor, bool atEnd, bool active)
{
    auto m = Motors::GetIt();
    Motors::MotorRoles role = m->mapVirtToRole(motor);
    
    if (!isMotorRolePumpStage(role))
        return;
    
    int motorNum = role - Motors::RoleStageX;
    if (motorCurrentStages[motorNum] == 0)
    {
        if (active) return;
        
        motorCurrentStages[motorNum] = 1;
        int dist = currentDistance[motorNum];
        int speed = m->getMotorMaxSpeed(motor);
        if (xyStageLimit[motorNum])
            m->goUp(motor, speed, dist);
        else
            m->goDown(motor, speed, dist);
        return;
    }
    
    if (motorCurrentStages[motorNum] == 1)
    {
        if (active) return;
        
        motorCurrentStages[motorNum] = -1;
        if (startMotorBtn->isOn)
            startMotorBtn->change_state_man(false);
    }
}

bool PumpsStage::isMotorRolePumpStage(Motors::MotorRoles role)
{
    return (role == Motors::RoleStageX || role == Motors::RoleStageY);
}

void PumpsStage::start(bool start)
{
    int distId = 0;
    if (radioSelectedConfLine[1]->isChecked())
        distId = 1;
    
    startMotor(start, currentMotorRole, distId);
}

void PumpsStage::startMotor(bool start, Motors::MotorRoles role, int distId)
{
    auto m = Motors::GetIt();
    
    if (!m->isMotorExist(role))
        return;
    
    auto motor = m->getRoleMotor(role);
    int motorNum = role - Motors::RoleStageX;
    if (start)
    {
        currentDistance[motorNum] = xyStagePosition[motorNum][distId];
        motorCurrentStages[motorNum] = 0;
        int speed = m->getMotorMaxSpeed(motor);
        if (xyStageLimit[motorNum])
            m->goDownE(motor, speed);
        else
            m->goUpE(motor, speed);
    }
    else
    {
        motorCurrentStages[motorNum] = -1;
        if (startMotorBtn->isOn)
            startMotorBtn->change_state_man(false);
        m->goStop(motor);
    }
}

void PumpsStage::setActiveMotor(Motors::MotorRoles role)
{
    this->setTitle(title + Motors::axes_names[role]);
    
    int motorNum = role - Motors::RoleStageX;
    currentMotorNum = motorNum;
    currentMotorRole = role;
    
    limitterCb->blockSignals(true);
    limitterCb->setChecked(xyStageLimit[motorNum]);
    limitterCb->blockSignals(false);
    
    for (int i=0; i<2; i++)
    {
        confLinesValuesGui[i]->blockSignals(true);
        confLinesValuesGui[i]->setValue(xyStagePosition[motorNum][i]);
        confLinesValuesGui[i]->blockSignals(false);
    }
}

void PumpsStage::startMoveStages(bool on)
{
    int distId = !on;
    startMotor(true, Motors::RoleStageX, distId);
    startMotor(true, Motors::RoleStageY, distId);
}
